# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details
from __future__ import print_function
from cubical.machines.parameterised_machine import ParameterisedGains
import numpy as np
from numpy.ma import masked_array
import cubical.kernels
from cubical.tools import logger

log = logger.getLogger("slopes")

def _normalize(x, dtype):
    """
    Helper function: normalizes array to [0,1] interval.
    """

    if len(x) > 1:
        return ((x - x[0]) / (x[-1] - x[0])).astype(dtype)
    elif len(x) == 1:
        return np.zeros(1, dtype)
    else:
        return x


import builtins
try:
    builtins.profile
except AttributeError:
    # No line profiler, provide a pass-through version
    def profile(func): return func
    builtins.profile = profile


class PhaseSlopeGains(ParameterisedGains):
    """
    This class implements the diagonal phase-only parameterised slope gain machine.
    """

    def __init__(self, label, data_arr, ndir, nmod, chunk_ts, chunk_fs, chunk_label, options):
        """
        Initialises a diagonal phase-slope gain machine.

        Args:
            label (str):
                Label identifying the Jones term.
            data_arr (np.ndarray):
                Shape (n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing observed
                visibilities.
            ndir (int):
                Number of directions.
            nmod (nmod):
                Number of models.
            chunk_ts (np.ndarray):
                Times for the data being processed.
            chunk_fs (np.ndarray):
                Frequencies for the data being processsed.
            options (dict):
                Dictionary of options.
        """
        ParameterisedGains.__init__(self, label, data_arr, ndir, nmod,
                                    chunk_ts, chunk_fs, chunk_label, options)

        self.slope_type = options["type"]
        self.n_param = 3 if self.slope_type == "tf-plane" else 2
        self._estimate_delays = options["estimate-delays"]
        self._pin_slope_iters = options["pin-slope-iters"]

        if self._pin_slope_iters > self.maxiter:
            raise ValueError("Slope pinning iterations is greater than the maximum "
                             "number of iterations. Please check term-iters, max-iters "
                             "and pin-slope-iters.")

        if (self.slope_type != "f-slope") and (self._pin_slope_iters != 0):
            raise ValueError("Slope pinning is not supported for modes other than "
                             "f-slope. Please ensure that pin-slope-iters is zero.")

        self.param_shape = [self.n_dir, self.n_timint, self.n_freint,
                            self.n_ant, self.n_param, self.n_cor, self.n_cor]

        # This needs to change - we want to initialise these slope params
        # from the fourier transform along the relevant axis.

        self.slope_params = np.zeros(self.param_shape, dtype=self.ftype)
        self.posterior_slope_error = None

        self.chunk_ts = _normalize(chunk_ts, self.ftype)
        self.chunk_fs = _normalize(chunk_fs, self.ftype)

        if self.slope_type == "tf-plane":
            self.slope = cubical.kernels.import_kernel("tf_plane")
            self._labels = dict(phase=0, delay=1, rate=2)
        elif self.slope_type == "f-slope":
            self.slope = cubical.kernels.import_kernel("f_slope")
            self._labels = dict(phase=0, delay=1)
            if self._estimate_delays:
                # Select all baselines containing the reference antenna.

                ref_ant_data = data_arr[:, :, :, self.ref_ant]

                # Average over time solution intervals. This should improve SNR,
                # but is not always necesssary.

                interval_data = np.add.reduceat(ref_ant_data, self.t_bins, 1)
                interval_smpl = np.add.reduceat(ref_ant_data != 0, self.t_bins, 1)
                selection = np.where(interval_smpl)
                interval_data[selection] /= interval_smpl[selection]

                bad_bins = np.add.reduceat(interval_data, self.f_bins, 2)

                # FFT the data along the frequency axis. TODO: Need to consider
                # what happens if we solve for few delays across the band. As there
                # is no guarantee that the band will be perfectly split, this may
                # need to be a loop over frequency solution intervals.

                pad_factor = options["delay-estimate-pad-factor"]
                
                for i in range(self.n_freint):
                    edges = self.f_bins + [None]
                    slice_fs = self.chunk_fs[edges[i]:edges[i+1]]

                    slice_data = interval_data[:, :, edges[i]:edges[i+1]]

                    slice_nchan = slice_data.shape[2]

                    fft_data = np.abs(np.fft.fft(slice_data, n=slice_nchan*pad_factor, axis=2))
                    fft_data = np.fft.fftshift(fft_data, axes=2)

                    # Convert the normalised frequency values into delay values.

                    delta_freq = slice_fs[1] - slice_fs[0]
                    fft_freq = np.fft.fftfreq(slice_nchan*pad_factor, delta_freq)
                    fft_freq = np.fft.fftshift(fft_freq)

                    # Find the delay value at which the FFT of the data is
                    # maximised. As we do not pad the values, this only a rough
                    # approximation of the delay. We also reintroduce the
                    # frequency axis for consistency.

                    delay_est_ind = np.argmax(fft_data, axis=2)
                    delay_est_ind = np.expand_dims(delay_est_ind, axis=2)

                    delay_est = fft_freq[delay_est_ind]
                    delay_est[...,(0,1),(1,0)] = 0
                    
                    # Check for bad data points (bls missing across all channels)
                    # and set their estimates to zero.

                    selection = np.where(bad_bins[:, :, i:i+1] == 0)
                    delay_est[selection] = 0

                    # Zero the off diagonals and take the negative delay values -
                    # this is necessary as we technically get the delay
                    # corresponding to the conjugate term.

                    self.slope_params[..., i:i+1, :, 1, :, :] = -1*delay_est
                    self.slope_params[..., (1,0), (0,1)] = 0
                    
                    log(1).print("{}: slope estimates are {}, {}".format(chunk_label,
                                        " ".join(map(str, self.slope_params[..., i:i+1, :, 1, 0, 0])),
                                        " ".join(map(str, self.slope_params[..., i:i+1, :, 1, 1, 1]))
                                         ))


        elif self.slope_type == "t-slope":
            self.slope = cubical.kernels.import_kernel("t_slope")
            self._labels = dict(phase=0, rate=1)
        else:
            raise RuntimeError("unknown machine type '{}'".format(self.slope_type))

        self.slope.construct_gains(self.slope_params, self.gains,
                                   self.chunk_ts, self.chunk_fs, self.t_int,
                                   self.f_int)

        # kernel used in solver is diag-diag in diag mode, else uses full kernel version
        if options.get('diag-data') or options.get('diag-only'):
            self.kernel_solve = cubical.kernels.import_kernel('diag_phase_only')
        else:
            self.kernel_solve = cubical.kernels.import_kernel('phase_only')

        self._jhr0 = self._gerr = None

    @property
    def has_converged(self):
        """ Returns convergence status. """

        condition_1 = self.converged_fraction >= self.min_quorum
        condition_2 = self.iters >= self.maxiter
        condition_3 = self.iters > self._pin_slope_iters

        return (condition_1 or condition_2) and condition_3

    @classmethod
    def determine_diagonality(cls, options):
        return True

    @classmethod
    def get_full_kernel(cls, options, diag_gains):
        from .phase_diag_machine import PhaseDiagGains
        return PhaseDiagGains.get_full_kernel(options, diag_gains)

    @staticmethod
    def exportable_solutions():
        """ Returns a dictionary of exportable solutions for this machine type. """

        exportables = ParameterisedGains.exportable_solutions()

        exportables.update({
            "phase": (0., ("dir", "time", "freq", "ant", "corr")),
            "delay":  (0., ("dir", "time", "freq", "ant", "corr")),
            "rate":   (0., ("dir", "time", "freq", "ant", "corr")),
            "phase.err": (0., ("dir", "time", "freq", "ant", "corr")),
            "delay.err": (0., ("dir", "time", "freq", "ant", "corr")),
            "rate.err": (0., ("dir", "time", "freq", "ant", "corr")),
        })

        return exportables

    def get_inverse_gains(self):
        """Returns inverse gains and inverse conjugate gains. For phase-only, conjugation is inverse"""
        gh = self.get_conj_gains()
        return gh, self.gains, 0

    def importable_solutions(self, grid0):
        """ Returns a dictionary of importable solutions for this machine type. """
        # Note that complex gain (as a derived parameter) is exported, but not imported
        # defines solutions we can import from
        return { label: dict(dir=grid0['dir'] if self.dd_term else [0], ant=grid0['ant'], corr=grid0['corr'], **self.interval_grid)
                 for label in self._labels.keys() }

    def export_solutions(self):
        """ Saves the solutions to a dict of {label: solutions,grids} items. """

        solutions = ParameterisedGains.export_solutions(self)

        for label, num in self._labels.items():
            solutions[label] = masked_array(self.slope_params[...,num,(0,1),(0,1)]), self.interval_grid
            if self.posterior_slope_error is not None:
                solutions[label+".err"] = masked_array(self.posterior_slope_error[..., num, :]), self.interval_grid

        
        with np.printoptions(precision=3, linewidth=100000, threshold=10000): 
            log(1).print("{}: slope solutions are {}, {}".format(self.chunk_label,
                            self.slope_params[..., :, :, 1, 0, 0],
                            self.slope_params[..., :, :, 1, 1, 1]))

        return solutions

    def import_solutions(self, soldict):
        """
        Loads solutions from a dict.

        Args:
            soldict (dict):
                Contains gains solutions which must be loaded.
        """

        # Note that this is inherently very flexible. For example, we can init from a solutions
        # table which only has a "phase" entry, e.g. one generated by a phase_only solver (and the
        # delay will then be left at zero).

        loaded = False
        for label, num in self._labels.items():
            value = soldict.get(label)
            if value is not None:
                self.slope_params[...,num,(0,1),(0,1)] = value
                loaded = True

        if loaded:
            self.restrict_solution(self.slope_params)
            self.slope.construct_gains(self.slope_params, self.gains,
                                           self.chunk_ts, self.chunk_fs, self.t_int, self.f_int)
            self._gains_loaded = True

    def compute_js(self, obser_arr, model_arr):
        """
        This function computes the J\ :sup:`H`\R term of the GN/LM method.

        Args:
            obser_arr (np.ndarray):
                Shape (n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing the
                observed visibilities.
            model_arr (np.ndrray):
                Shape (n_dir, n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing the
                model visibilities.

        Returns:
            np.ndarray:
                J\ :sup:`H`\R
        """

        n_dir, n_tim, n_fre, n_ant, n_cor, n_cor = self.gains.shape

        gh = self.get_conj_gains()
        jh = self.get_new_jh(model_arr)

        self.slope.compute_jh(model_arr, self.gains, jh, 1, 1)

        jhr1 = self.get_new_jhr()

        r = self.get_obs_or_res(obser_arr, model_arr)

        # use appropriate phase-only kernel (with 1,1 intervals) to compute inner JHR
        self.kernel_solve.compute_jhr(gh, jh, r, jhr1, 1, 1)

        jhr1 = jhr1.imag

        jhr_shape = [n_dir, self.n_timint, self.n_freint, n_ant, self.n_param, n_cor, n_cor]

        if self._jhr0 is None:
            self._jhr0 = self.slope.allocate_param_array(jhr_shape, dtype=jhr1.dtype, zeros=True)
        else:
            self._jhr0.fill(0)

        self.slope.compute_jhr(jhr1, self._jhr0, self.chunk_ts, self.chunk_fs, self.t_int, self.f_int)

        return self._jhr0, self.jhjinv, 0

    @property
    def dof_per_antenna(self):
        """This property returns the number of real degrees of freedom per antenna, per solution interval"""
        if self.slope_type=="tf-plane":
            return 6
        elif self.slope_type=="f-slope":
            return 4
        elif self.slope_type=="t-slope":
            return 4

    def implement_update(self, jhr, jhjinv):

        # variance of slope parms is diagonal of jhjinv
        diag = (0,2) if self.n_param == 2 else (0,3,5)   # weird numbering to get diagonal elements
        var_slope = jhjinv[...,(0,1),(0,1)].real[...,diag,:]
        if self.posterior_slope_error is None:
            self.posterior_slope_error = np.sqrt(var_slope)
        else:
            np.sqrt(var_slope, out=self.posterior_slope_error)
            
        #import ipdb; ipdb.set_trace()
        with np.printoptions(precision=3, linewidth=100000, threshold=10000): 
            log(1).print("Iteration {}".format(self.iters))
            log(1).print("slope variance X: ", var_slope[0,:,0,:,1,0].max(axis=0))
            log(1).print("slope variance Y: ", var_slope[0,:,0,:,1,1].max(axis=0))
            
        # variance of gain is sum of slope parameter variances
        if self._gerr is None:
            self._gerr = var_slope.sum(axis=-2)
        else:
            var_slope.sum(axis=-2, out=self._gerr)

        np.sqrt(self._gerr, out=self._gerr)
        if self.posterior_gain_error is None:
            self.posterior_gain_error = np.zeros_like(self.gains, dtype=float)
        self._interval_to_gainres(self._gerr[...,0], 1, out=self.posterior_gain_error[...,0,0])
        self._interval_to_gainres(self._gerr[...,0], 1, out=self.posterior_gain_error[...,1,1])

        update = self.init_update(jhr)

        self.slope.compute_update(jhr, jhjinv, update)
        
        with np.printoptions(precision=3, linewidth=100000, threshold=10000): 
            log(1).print("proposed slope update X: ", update[0,:,0,:,1,0,0].max(axis=0))
            log(1).print("proposed slope update Y: ", update[0,:,0,:,1,1,1].max(axis=0))
            log(1).print("proposed phase update X (deg): ", update[0,:,0,:,0,0,0].max(axis=0)*180/np.pi)
            log(1).print("proposed phase update Y (deg): ", update[0,:,0,:,0,1,1].max(axis=0)*180/np.pi)

        # Pin the delay component if self._pin_slope_iters is set. Note that
        # this is likely incorrect for delay+rate calibration.
        if self.iters < self._pin_slope_iters:
            update[..., 1, :, :] = 0

        # It is safer to average every iteration when using a phase only solver.
        if self.iters%1 == 0:
            update *= 0.5
        
        # suppress slope updates for the first 10 iterations, to let phase settle    
        #if self.iters < 10:
        #    update *= 0.01
        #    log(1).print("update damped")

        self.slope_params += update

        self.restrict_solution(self.slope_params)

        # Need to turn updated parameters into gains.

        self.slope.construct_gains(self.slope_params, self.gains, self.chunk_ts, self.chunk_fs, self.t_int, self.f_int)

        # raise flag so updates of G^H and G^-1 are computed
        self._gh_update = self._ghinv_update = True


    def restrict_solution(self, slope_params):
        """
        Restricts the solution by invoking the inherited restrict_solution method and applying
        any machine specific restrictions.
        """

        # ParameterisedGains.restrict_solution(self)
        # These need to be set here as we do not call the interval restrict solutions -
        # out solutions are the parameters and not the gains themselves.
        self._gh_update = self._ghinv_update = True

        if self.ref_ant is not None:
            # complicated slice :) we take the 0,0 phase offset of the reference antenna,
            # and subtract that from the phases of all other antennas and elements
            ref_slice = slice(self.ref_ant, self.ref_ant+1, 1)
            ref_params = slope_params[:,:,:,ref_slice,:,0,0]
            slope_params[:,:,:,:,:,(0,1),(0,1)] -= ref_params[..., None]

        for idir in self.fix_directions:
            slope_params[idir, ...] = 0

    def precompute_attributes(self, data_arr, model_arr, flags_arr, inv_var_chan):
        """
        Precompute (J\ :sup:`H`\J)\ :sup:`-1`, which does not vary with iteration.

        Args:
            model_arr (np.ndarray):
                Shape (n_dir, n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing
                model visibilities.
        """
        ParameterisedGains.precompute_attributes(self, data_arr, model_arr, flags_arr, inv_var_chan)

        jhj1_shape = [self.n_dir, self.n_tim, self.n_fre, self.n_ant, 2, 2]

        jhj1 = self.slope.allocate_gain_array(jhj1_shape, dtype=self.dtype, zeros=True)

        # use appropriate phase-only kernel (with 1,1 intervals) to compute inner JHJ
        self.kernel_solve.compute_jhj(model_arr, jhj1, 1, 1)

        blocks_per_inverse = 6 if self.slope_type=="tf-plane" else 3

        jhj_shape = [self.n_dir, self.n_timint, self.n_freint, self.n_ant, blocks_per_inverse, 2, 2]

        jhj = self.slope.allocate_param_array(jhj_shape, dtype=self.ftype, zeros=True)

        self.slope.compute_jhj(jhj1.real, jhj, self.chunk_ts, self.chunk_fs, self.t_int, self.f_int)

        self.jhjinv = np.zeros_like(jhj)

        self.slope.compute_jhjinv(jhj, self.jhjinv, self.eps)




# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from multiprocessing.sharedctypes import Value

from DDFacet.compatibility import range
import DDFacet.Data.ClassFITSBeam as DDFacetFITSBeam
import DDFacet.Data.ClassMS as DDFacetClassMS
import numpy
import os
import os.path
import sys
import pyrap.tables
import numpy as np

from cubical.tools import logger, ModColor
log = logger.getLogger("FITSBeamInterpolator")

dm = pyrap.measures.measures()
dq = pyrap.quanta

# This a list of the Stokes enums (as defined in casacore header measures/Stokes.h)
# These are referenced by the CORR_TYPE column of the MS POLARIZATION subtable.
# E.g. 5,6,7,8 corresponds to RR,RL,LR,LL
MS_STOKES_ENUMS = DDFacetFITSBeam.MS_STOKES_ENUMS
# set of circular correlations
CIRCULAR_CORRS = DDFacetFITSBeam.CIRCULAR_CORRS
# set of linear correlations
LINEAR_CORRS = DDFacetFITSBeam.LINEAR_CORRS

class DDFacetMSLite:
    def __init__ (self,
                  field_centre,
                  chunk_ctr_frequencies,
                  chunk_channel_widths,
                  correlation_ids,
                  station_positions,
                  station_names):
        """
            Standin object for the ClassMS visibility reader of DDFacet
            field_centre: radian ra dec coordinate of the original field phase centre
            chunk_ctr_frequencies: center channel frequencies (Hz) of chunk to predict E terms for
            chunk_channel_widths: channel widths of chunk channels
            correlation_ids: hand labels per correlation term as defined in casacore Stokes.h
            station_positions: ECEF coordinates for stations (na, 3) array
            station_names: Station names
        """
        self.ChanFreq = chunk_ctr_frequencies
        self.OriginalRadec = field_centre
        self.ChanWidth = chunk_channel_widths
        self.CorrelationNames = list(map(lambda x: MS_STOKES_ENUMS[x], 
                                         sorted(correlation_ids)))
        self.StationNames = station_names
        self.na = len(station_names)
        self.StationPos = station_positions
        self.rarad, self.decrad = self.OriginalRadec
        if len(self.StationNames) != len(self.StationPos):
            raise ValueError("StationNames should be the length of StationPos")
        
    def radec2lm_scalar(self, *args, **kwargs):
        return DDFacetClassMS.ClassMS.radec2lm_scalar(self, *args, **kwargs)
        
class FITSBeamInterpolator (object):
    def __init__ (self, 
                  field_centre,
                  chunk_ctr_frequencies,
                  chunk_channel_widths,
                  correlation_ids,
                  station_positions,
                  station_names,
                  opts):
        """
            Proxy object for DDFacet.ClassFITSBeam
            field_centre: radian ra dec coordinate of the original field phase centre
            chunk_ctr_frequencies: center channel frequencies (Hz) of chunk to predict E terms for
            chunk_channel_widths: channel widths of chunk channels
            correlation_ids: hand labels per correlation term as defined in casacore Stokes.h
            station_positions: ECEF coordinates for stations (na, 3) array
            station_names: Station names
            opts: beam opts as defined in Parset
        """
        self.__ms = DDFacetMSLite(field_centre,
                                  chunk_ctr_frequencies,
                                  chunk_channel_widths,
                                  correlation_ids,
                                  station_positions,
                                  station_names)
        self.__opts = opts
        self.__vbinterpolator = DDFacetFITSBeam.ClassFITSBeam(self.__ms, self.__opts)
        for method in ('evaluateBeam', 'getFreqDomains', 'getBeamSampleTimes'):
            setattr(self, method, getattr(self.__vbinterpolator, method))






from __future__ import print_function
try:
    from DDFacet.Imager import ModModelMachine as MMM
    from DDFacet.ToolsDir.ModToolBox import EstimateNpix
except ImportError:
    raise ImportError("Cannot import DDFacet")

from regions import DS9Parser
import numpy as np
from .geometry import BoundingConvexHull, BoundingBox, BoundingBoxFactory
from cubical.tools import logger, ModColor
log = logger.getLogger("DDFacetSim")

class DicoSourceProvider(object):
    __cached_model_machines = {}
    __cached_providers = {}

    @classmethod
    def __cachename_compute(cls, diconame, regname):
        return "src_model_{0:s}_@_{1:s}".format(diconame,
                                                regname if regname is not None else "die")

    def __init__(self,
                 fn,
                 phasedir,
                 facet_padding_factor,
                 max_facet_size,
                 min_nfacet_per_axis,
                 clustercat=None):
        self.__dicofn = fn

        if fn not in DicoSourceProvider.__cached_model_machines:
            self.__dicomodel = MMM.ClassModModelMachine().GiveInitialisedMMFromFile(fn)
            DicoSourceProvider.__cached_model_machines[fn] = self.__dicomodel
        else:
            log(2).print("Reusing previously initialized DDFacet model machine '{0:s}'".format(fn))
            self.__dicomodel = DicoSourceProvider.__cached_model_machines[fn]

        # assume gd is stored in the dico... this is only true for ddf versions 0.4.1 and beyond
        self.__pxscale = self.__dicomodel.GD["Image"]["Cell"] # in arcseconds

        self.__nchan, self.__npol, self.__ny, self.__nx = self.__dicomodel.ModelShape
        assert self.__ny == self.__nx # check cyril's model is square
        assert self.__nx % 2 == 1 # must be odd for the logic downstairs to work

        # regions relative to the unpadded images
        req_npx_img = self.__dicomodel.GD["Image"]["NPix"]
        nx_image, nx_image_padded = EstimateNpix(req_npx_img,
                                                 Padding=self.__dicomodel.GD["Facets"]["Padding"])
        assert nx_image == self.__nx # the predicted image size better correspond to dicomodels shape
        self.__padded_nx = self.__padded_ny = nx_image_padded
        self.__phasedir = phasedir

        cachename = DicoSourceProvider.__cachename_compute(fn, clustercat)
        if cachename not in DicoSourceProvider.__cached_providers:
            log.info("Initializing new source provider for DDFacet model '{0:s}' into regions specified by '{1:s}'.".format(
                fn, clustercat if clustercat is not None else 'die'))
            self.__clustercat = self.__read_regions_file(clustercat, 
                                                         facet_padding_factor, 
                                                         max_facet_size,
                                                         min_nfacet_per_axis)
            DicoSourceProvider.__cached_providers[cachename] = self.__clustercat
            log(2).print("initialization sequence of source provider '{0:s}' (regions '{1:s}') completed".format(
                    fn, clustercat if clustercat is not None else 'die'))
        else:
            self.__clustercat = DicoSourceProvider.__cached_providers[cachename]
            log(2).print("reused previous initialization of source provider '{0:s}' (regions '{1:s}')".format(
                fn, clustercat if clustercat is not None else 'die'))

        self.__current_direction = 0
        self.__degridcube = None
        self.__degridfreqs = None
        self.__padding_factor = facet_padding_factor

    def __str__(self):
        return self.__dicofn

    @property
    def subregion_count(self):
        return len(self.__clustercat[self.__current_direction])

    @property
    def pixel_scale(self):
        return self.__pxscale

    @property
    def degrid_cube_shape(self):
        if self.__degridcube is None:
            raise RuntimeError("Degrid frequencies not set yet. Please set frequencies before enquiring about the degrid cube")
        return self.__degridcube.shape

    def __read_regions_file(self, fn, padding_factor, max_size, min_nfacet_per_axis):
        """ Reads a ds9 region file and sets up a grid of sub clusters (aka. facets) for each hull read out of
            the file. Each of the sub clusters are padded to a padding factor as determined by
            DDFacet's EstimateNpix. The number of sub clusters is determined from the maximum
            size of such sub cluster.

            If the fn is None it is assumed that the user intends to split the full sky up into a single
            direction with each sub cluster contributing to this single direction (i.e. regular
            tesselation of the sky into square facets).

            Subclusters are guaranteed to be odd sized and square

            return dictionary of directions, each entry containing a list of sub regions
        """
        clusters = []
        if fn is not None: # dde case
            with open(fn) as f:
                parser = DS9Parser(f.read())
                for regi, reg in enumerate(parser.shapes):
                    coords = list(map(int, [c.value for c in reg.coord]))
                    assert len(coords) % 2 == 0, "Number of region coords must be multiple of 2-tuple"
                    coords = np.array(coords).reshape([len(coords) // 2, 2])
                    clusters.append(BoundingConvexHull(coords,
                                                       name="DDE_REG{0:d}".format(regi + 1)))
        else: # die case
            clusters = [BoundingBox(0,
                                    self.__nx - 1,
                                    0,
                                    self.__ny - 1,
                                    name="die",
                                    check_mask_outofbounds=True)]
        log(2).print("\tInitialized bounding boxes for regions. There are {0:d} region(s)".format(len(clusters)))
        # now create axis aligned bounding boxes for each of the regions
        # and further split them to the maximum permitted facet size
        clusters = [BoundingBoxFactory.AxisAlignedBoundingBox(c, check_mask_outofbounds=False) for c in clusters]
        def __split_regular_region(reg, max_size):
            if max_size < 0:
                raise ValueError("Expected positive value for min_size")
            reg_size_deg = np.max(np.array(reg.box_npx) * self.pixel_scale / 3600.0)
            nsplit = max(1, max(min_nfacet_per_axis, int(np.ceil(reg_size_deg / max_size))))
            return BoundingBoxFactory.SplitBox(reg, nsubboxes=nsplit, check_mask_outofbounds=True)
        log(2).print("\tSplitting regions into facetted regions, with maximum unpadded size of {0:.2f} degrees per facet".format(max_size))
        clusters = [aasubreg for aareg in map(lambda reg: __split_regular_region(reg, max_size), clusters) 
                    for aasubreg in aareg]
        clusters = list(map(lambda reg: BoundingBoxFactory.AxisAlignedBoundingBox(reg, square=True, check_mask_outofbounds=False), clusters))
        
        def __pad_cluster(c, padding_factor):
            npx,_ = c.box_npx # square facet at this point
            # this returns an odd npix:
            npixunpadded, npixpadded = EstimateNpix(npx, Padding=padding_factor)
            return BoundingBoxFactory.PadBox(c, npixpadded, npixpadded, check_mask_outofbounds=False)
        log(2).print("\tPadding all facets by a minimum factor of {0:.2f}x".format(padding_factor))
        clusters = list(map(lambda c: __pad_cluster(c, padding_factor), clusters))
        log.debug("\tNormalizing regional weights")
        BoundingConvexHull.normalize_masks(clusters)
        log(2).print("\tCaching regional weight maps for future predicts")
        list(map(lambda x: x.mask, clusters)) # cache mask
        dirs = {} 
        for c in clusters:
            dirs[c.name] = dirs.get(c.name, []) + [c]
        return dirs

    @property
    def _cluster_keys(self):
        """ returns list of cluster name keys if there are any clusters
            otherwise ["die"] if no dE clusters were specified upon construction
        """
        return ["die"] if self.__clustercat is None else \
            self.__clustercat.keys()
    
    def set_direction(self, v):
        """ sets the cluster (direction) index to predict from """
        if v not in self._cluster_keys:
            raise KeyError("Unknown direction '{0:s}' for model '{1:s}'".format(str(v), str(self)))
        self.__current_direction = v
    
    @property
    def _nclus(self):
        return len(self.__clustercat.keys())

    @property
    def direction(self):
        return self.__current_direction

    @direction.setter
    def direction(self, v):
        self.set_direction(v)

    @classmethod
    def crosspattern(cls, model_image, nterms = 7, first_position_from_centre=0.05, last_position_from_edge=0.8):
        if model_image.shape[2] != model_image.shape[3]:
            raise ValueError("Expected square grid")
        model_image[...] = 0
        for v in np.linspace(model_image.shape[2]//2 * first_position_from_centre, 
                             model_image.shape[2]//2 * last_position_from_edge, nterms):
            model_image[:, :, model_image.shape[2]//2 - int(v), model_image.shape[3]//2 - int(v)] = 1.0
            model_image[:, :, model_image.shape[2]//2 + int(v), model_image.shape[3]//2 + int(v)] = 1.0
            model_image[:, :, model_image.shape[2]//2 + int(v), model_image.shape[3]//2 - int(v)] = 1.0
            model_image[:, :, model_image.shape[2]//2 - int(v), model_image.shape[3]//2 + int(v)] = 1.0
        return model_image


    def set_frequency(self, freqs):
        """ Sets the model prediction frequencies. This initializes the model degridding bands """
        self.__degridfreqs = freqs.copy()
        self.__degridcube = self.__dicomodel.GiveModelImage(FreqIn=self.__degridfreqs)
        #self.__degridcube = DicoSourceProvider.crosspattern(self.__degridcube) #widefield debug

    def get_direction_npix(self, subregion_index=0):
        if subregion_index >= len(self.__clustercat[self.__current_direction]):
            raise IndexError("Index {0:d} is out of bounds for region {1:s} which contains {2:d} subregions".format(
                subregion_index, self.__current_direction, len(self.__clustercat[self.__current_direction])))
        return self.__clustercat[self.__current_direction][subregion_index].box_npx[0]

    def get_direction_pxoffset(self, subregion_index=0):
        """ get the pixel offset to the centre of the facet """
        if subregion_index >= len(self.__clustercat[self.__current_direction]):
            raise IndexError("Index {0:d} is out of bounds for region {1:s} which contains {2:d} subregions".format(
                subregion_index, self.__current_direction, len(self.__clustercat[self.__current_direction])))

        cluster = self.__clustercat[self.__current_direction][subregion_index]
        ctr = cluster.centre
        offset = (ctr[::-1] - np.array([self.__ny//2, self.__nx//2]))[::-1]
        offset[0] = -offset[0] #RA increases to the left
        return offset

    def get_degrid_model(self, subregion_index=0):
        """ get a facet from model cube for the direction previously set by set_direction """
        if self.__degridcube is None:
            raise ValueError("Set predict frequencies before running this method")
        if subregion_index >= len(self.__clustercat[self.__current_direction]):
            raise IndexError("Index {0:d} is out of bounds for region {1:s} which contains {2:d} subregions".format(
                subregion_index, self.__current_direction, len(self.__clustercat[self.__current_direction])))
        reg_data, reg_extents = BoundingConvexHull.regional_data(self.__clustercat[self.__current_direction][subregion_index],
                                                                 np.swapaxes(self.__degridcube, 2, 3)[:, :, :, ::-1])
        # cyril's model is transposed and flipped
        reg_data[...] = reg_data[:, :, ::-1, ::-1]
        return reg_data



# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details
from __future__ import print_function
from six import add_metaclass
from abc import ABCMeta, abstractmethod, abstractproperty
import numpy as np
from numpy.ma import masked_array
import traceback

from cubical import param_db
from cubical.database.casa_db_adaptor import casa_db_adaptor

from cubical.tools import logger, ModColor

log = logger.getLogger("gain_machine")


@add_metaclass(ABCMeta)
class MasterMachine:
    """
    This is a base class for all solution machines. It is completely generic and lays out the basic
    requirements for all machines.
    
    It also provides a Factory class that takes care of creating machines and interfacing with 
    solution tables on disk.
    """
    __user_warnings = []

    def __init__(self, jones_label, data_arr, ndir, nmod, times, freqs, chunk_label, options, diagonal=None):
        """
        Initializes a gain machine.
        
        All supplied arguments are copied to attributes of the same name. In addition, the following
        attributes or properties are populated. These may be considered part of the official machine interface.
            
            - solvable:
                from options['solvable'], default False
            - dd_term:
                from options['dd-term'], default False
            - n_dir, n_mod, n_tim, n_fre, n_ant, n_cor:
                problem dimensions
            - dtype:
                complex dtype of data/model (complex64 or complex128)
            - ftype:
                corresponding float dtype (float32 or float64)
            - iters:
                iteration counter, reset to 0.
            - maxiters:
                from options['max-iters'], default 0
        
        Args:
            jones_label (str):
                Label identifying the Jones term.
            data_arr (np.ndarray): 
                Shape (n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing observed 
                visibilities. 
            ndir (int):
                Number of directions.
            nmod (nmod):
                Number of models.
            times (np.ndarray):
                Times for the data being processed.
            freqs (np.ndarray):
                Frequencies for the data being processsed.
            chunk_label (str):
                Label of the data chunk being processed, for messages
            options (dict):
                Dictionary of options.
            diagonal (bool or None):
                Set to False or True or False if the gains are (non)diagonal. If None, calls the
                determine_diagonality() class method instead

        """
        import cubical.kernels
        self.generics = cubical.kernels.import_kernel('generics')

        self._jones_label = jones_label
        self.chunk_label = chunk_label
        self.times = times
        self.freqs = freqs
        self.options = options
        self._is_diagonal = self.determine_diagonality(options)
        self._allocate_vis_array, self._allocate_flag_array, self._allocate_gain_array = self.determine_allocators(options)

        self.solvable = options.get('solvable')
        self._dd_term = options.get('dd-term')
        self._maxiter = options.get('max-iter', 0)

        self._diag_only = options.get('diag-only')
        self._offdiag_only = options.get('offdiag-only')

        self._prop_flags = options.get('prop-flags', 'default')

        self._epsilon = options.get('epsilon')
        self._delta_chi = options.get('delta-chi')

        self.n_dir, self.n_mod = ndir if self._dd_term else 1, nmod
        _, self.n_tim, self.n_fre, self.n_ant, self.n_ant, self.n_cor, self.n_cor = data_arr.shape

        self.dtype = data_arr.dtype
        self.ftype = data_arr.real.dtype

        self._iters = 0

    @property
    def jones_label(self):
        return self._jones_label

    @property
    def is_diagonal(self):
        """Returns true if this machine instance represents a diagonal gain term"""
        return self._is_diagonal

    @property
    def allocate_vis_array(self):
        """Returns visibility allocator function for this machine"""
        return self._allocate_vis_array

    @property
    def allocate_flag_array(self):
        """Returns flag allocator function for this machine"""
        return self._allocate_flag_array

    @property
    def allocate_gain_array(self):
        """Returns flag allocator function for this machine"""
        return self._allocate_gain_array

    @classmethod
    def determine_allocators(cls, options):
        """
        Class method. Given a machine class and a set of options, returns allocator
        functions appropriate to this machine type.

        Returns:
            tuple:
                - visibility array allocator
                - flag array allocator
                - gain array allocator
        """
        return NotImplementedError

    @classmethod
    def determine_diagonality(cls, options):
        """
        Class method. Given a machine class and a set of options, returns True if the machine is
        going to represent a diagonal-only gain term.
        """
        return NotImplementedError

    @property
    def dd_term(self):
        """This property is true if the machine represents a direction-dependent gain"""
        return self._dd_term

    @property
    def diag_only(self):
        """This property is true if the machine solves using diagonal visibilities only"""
        return self._diag_only

    @property
    def offdiag_only(self):
        """This property is true if the machine solves using off-diagonal visibilities only"""
        return self._offdiag_only

    @property
    def epsilon(self):
        return self._epsilon

    @property
    def delta_chi(self):
        return self._delta_chi

    @property
    def propagates_flags(self):
        return self._prop_flags == "always" or self._prop_flags == "any" or \
               (self._prop_flags == "default" and not self.dd_term)

    @property
    def propagates_anydir_flags(self):
        return self._prop_flags == "any"

    @property
    def maxiter (self):
        """This property gives the max number of iterations"""
        return self._maxiter

    @maxiter.setter
    def maxiter (self, value):
        """Sets max number of iterations"""
        self._maxiter = value

    @property
    def iters (self):
        """This property gives the current iteration counter."""
        return self._iters

    @iters.setter
    def iters(self, value):
        """Sets current iteration counter"""
        self._iters = value

    @abstractmethod
    def compute_js(self, obser_arr, model_arr):
        """
        This method is expected to compute (J\ :sup:`H`\J)\ :sup:`-1` and J\ :sup:`H`\R. 
        
        Args:
            model_arr (np.ndrray): 
                Shape (n_dir, n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing the 
                model visibilities.
            obser_arr (np.ndarray): 
                Shape (n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing the 
                observed visibilities.            

        
        Returns:
            3-element tuple
                
                - J\ :sup:`H`\R (np.ndarray)
                - (J\ :sup:`H`\J)\ :sup:`-1` (np.ndarray)
                - Count of flags raised (int)     
        """

        return NotImplementedError

    @abstractmethod
    def implement_update(self, jhr, jhjinv):
        """
        Internal method implementing a parameter update. The standard compute_update() implementation 
        calls compute_js() and _implement_update() at each iteration step.
        
        Args:
            jhr (np.ndarray):
                J\ :sup:`H`\R term
            jhjinv (np.ndarray):
                (J\ :sup:`H`\J)\ :sup:`-1` term
        """
        return NotImplementedError

    def compute_update(self, model_arr, obser_arr):
        """
        This method is expected to compute the parameter update. 
        
        The standard implementation simply calls compute_js() and implement_update(), but subclasses are free to 
        override.

        Args:
            model_arr (np.ndarray): 
                Shape (n_dir, n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing 
                model visibilities. 
            obser_arr (np.ndarray): 
                Shape (n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing observed 
                visibilities. 
        """
        jhr, jhjinv, flag_count = self.compute_js(obser_arr, model_arr)

        self.implement_update(jhr, jhjinv)

        return flag_count

    @abstractmethod
    def compute_residual(self, obser_arr, model_arr, resid_arr, require_full=True):
        """
        This method should compute the residual at the the full time-frequency resolution of the
        data. Must populate resid_arr with the values of the residual. Function signature must be 
        consistent with the one defined here.

        Args:
            obser_arr (np.ndarray): 
                Shape (n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing observed 
                visibilities. 
            model_arr (np.ndarray): 
                Shape (n_dir, n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing 
                model visibilities.
            resid_arr (np.ndarray):
                Shape (n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array in which to place the 
                residual values.
            require_full (bool):
                If True, a full 2x2 residual is required. If False, only the terms used in the solution
                (e.g. if a solution is diagonal, only the diagonals) are required.
        """

        return NotImplementedError

    @abstractmethod
    def apply_inv_gains(self, obser_arr, corr_vis=None, full2x2=True, direction=None):
        """
        This method should be able to apply the inverse of the gains associated with the gain
        machines to an array at full time-frequency resolution. Should populate an input array with
        the result or return a new array. Function signature must be consistent with the one defined
        here.

        Args:
            obser_arr (np.ndarray):
                Shape (n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing observed 
                visibilities.
            corr_vis (np.ndarray or None, optional):
                Shape (n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array to fill with the corrected 
                visibilities.
            full2x2 (bool):
                If True, gains should be applied to the full 2x2 matrix. If False, only the terms used in the solution
                (e.g. the diagonals) are required.
            direction (int or None):
                If None, only DI terms (in a chain) are applied. If set to a direction number,
                also applies DD terms for that direction.

        Returns:
            2-element tuple
                
                - Corrected visibilities (np.ndarray)
                - Flags raised (int)
        """
        
        return NotImplementedError

    @abstractmethod
    def apply_gains(self, model_arr, full2x2=True, dd_only=False):
        """
        This method should be able to apply the gains associated with the gain
        to an array at full time-frequency resolution. 
        Function signature must be consistent with the one defined here.

        Args:
            model_arr (np.ndarray):
                Shape (n_dir, n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing model visibilities.
            full2x2 (bool):
                If True, gains should be applied to the full 2x2 matrix. If False, only the terms used in the solution
                (e.g. the diagonals) are required.
            dd_only (bool):
                If True, only DD terms are applied, beginning with the leftmost. Not implemented for now.
        """

        return NotImplementedError

    @property
    def has_valid_solutions(self):
        """
        This property returns True if the gain machine has any valid solutions defined. (The opposite would be
        true if e.g. they were all flagged).
        """
        return False

    @property
    def num_solutions(self):
        """
        This property gives the number of solutions (e.g. intervals) defined by the machine
        """
        return 0

    @property
    def num_converged_solutions(self):
        """
        This property gives the number of currently converged solutions defined by the machine
        """
        return 0

    def update_equation_counts(self, unflagged):
        """
        Sets up equation counters and normalization factors. Sets up the following attributes:
            - eqs_per_record
                integer: gives the nominal number of equations per a visibility record, so e.g.
                would be 8 for a full 2x2 complex correlation matrix
            - eqs_per_tf_slot (np.ndarray):
                Shape (n_tim, n_fre) array containing a count of equations per time-frequency slot.
            - eqs_per_antenna (np.ndarray)
                Shape (n_ant, ) array containing a count of equations per antenna.

        Also sets up corresponding chi-sq normalization factors.

        Normally called form precompute_attributes() to set things up. Should also be called every time
        the flags change (otherwise chi-sq normalization will go off)

        Args:
            unflagged (np.ndarray):
                Shape (n_tim, n_fre, n_ant, n_ant) bool array indicating valid (not flagged) slots
        """

        # equations per visibility slot. We have two (real/imag, or normal and conjugate) per model
        # and per correlation product. This is just a normalization factor used in the bookkeeping.
        self.eqs_per_record = 2 * self.n_mod * self.n_cor
        # full 2x2 as opposed to diag-only or off-diag-only? Multiply by 2
        if not self.diag_only or self.offdiag_only:
            self.eqs_per_record *= self.n_cor

        # (n_ant) vector containing the number of valid equations per antenna (for any direction)

        self.eqs_per_antenna = np.sum(unflagged, axis=(0, 1, 2)) * self.eqs_per_record

        # (n_tim, n_fre) array containing number of valid equations for each time/freq slot

        self.eqs_per_tf_slot = np.sum(unflagged, axis=(-1, -2)) * self.eqs_per_record


        with np.errstate(invalid='ignore', divide='ignore'):
            self._chisq_tf_norm_factor = 1./self.eqs_per_tf_slot
        self._chisq_tf_norm_factor[self.eqs_per_tf_slot==0] = 0

        toteq = np.sum(self.eqs_per_tf_slot)
        self._chisq_norm_factor = 1./toteq if toteq else 0

        self._chisq_tf_norm_factor_0 = self._chisq_tf_norm_factor
        self._chisq_norm_factor_0 = self._chisq_norm_factor


    def compute_chisq(self, resid_arr, inv_var_chan, require_full=True):
        """
        Computes chi-squared statistics based on given residuals.
        Args:
            resid_arr (np.ndarray):
                Shape (n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing residuals.
            inv_var_chan (np.ndarray)
                Shape (nfreq,) array of 1/sigma^2 per channel

        Returns:
            3-tuple consisting of
            
             - chisq (np.ndarray):
                (n_tim, n_fre, n_ant) array of chi-squared per antenna and time/frequency slot
             - chisq_per_tf_slot  (np.ndarray):
                (n_tim, n_fre) array of chi-squared per time/frequency slot
             - chisq_tot (float):
                overall chi-squared value for the entire chunk
        """

        # Chi-squared is computed by summation over antennas, correlations and intervals. Sum over
        # time intervals, antennas and correlations first. Normalize by per-channel variance and
        # finally sum over frequency intervals.

        chisq = np.zeros(resid_arr.shape[1:4], np.float64)
        if self.diag_only:
            self.generics.compute_chisq_diag(resid_arr, chisq)
        elif self.offdiag_only:
            self.generics.compute_chisq_offdiag(resid_arr, chisq)
        else:
            self.generics.compute_chisq(resid_arr, chisq)

        # Normalize this by the per-channel variance.

        chisq *= inv_var_chan[np.newaxis, :, np.newaxis]

        # Collapse chisq to chi-squared per time-frequency slot and overall chi-squared

        chisq_per_tf_slot = np.sum(chisq, axis=-1) * self._chisq_tf_norm_factor

        chisq_tot = np.sum(chisq) * self._chisq_norm_factor

        return chisq, chisq_per_tf_slot, chisq_tot

    def precompute_attributes(self, data_arr, model_arr, flags_arr, inv_var_chan):
        """
        This method is called before starting a solution. The base version computes a variety of useful 
        parameters regarding the conditioning and degrees of freedom of the current time-frequency chunk. 
        Subclasses can redefine this to precompute other useful stuff (i.e. things that do not vary with 
        iteration).

        Args:
            data_arr (np.ndarray):
                Shape (n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing observed visibilities.
            model_arr (np.ndarray):
                Shape (n_dir, n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing model visibilities.
            flags_arr (np.ndarray):
                Shape (n_tim, n_fre, n_ant, n_ant) array containing integer flags
            inv_var_chan (np.ndarray)
                Shape (nfreq,) array of 1/sigma^2 per channel
                
        Returns:
            bool np.ndarray, shape (n_tim, n_fre, n_ant, n_ant), indicating the inverse of flags
        """
        unflagged = flags_arr==0
        self.update_equation_counts(unflagged)
        return unflagged

    @abstractmethod
    def check_convergence(self, min_delta_g):
        """
        This method should check the gain solutions for convergence.

        Args:
            min_delta_g (float):
                Threshold for the minimum change in the gains - convergence criterion.
        """

        return NotImplementedError

    @abstractmethod
    def flag_solutions(self, flag_arr, final=0):
        """
        This method allows the machine to flag gains solutions.
        When solving, it is called after each iteration (final=0), and then once again after convergence (final=1).
        When applying, it is called once after loading the solutions with final=-1
        
        This method can propagate the flags raised by the gain machine back into the data flags.
        
        Args:
            flag_arr (np.ndarray):
                Shape (n_tim, n_fre, n_ant, n_ant) array containing data flags. 
            final (int):
                -1 when loading (in apply-only mode), 0 while iterating, 1 after convergence.

        Returns:
            True if any new flags have been propagated to the data
        
        """
        return NotImplementedError

    @abstractmethod
    def num_gain_flags(self, mask=None, final=False):
        """
        This method returns the number of gains flagged, and the total number of gains.
        
        Args:
            mask (int):
                Flag mask to apply. If None, ~FL.MISSING is expected to be used
            final:
                If True, this is the final (post-solution) call, so return sum of all flags (if e.g. a chain)
 
        Returns:
            Tuple of two values
                - number of flagged gains
                - total number of gains

        """
        return NotImplementedError

    def next_iteration(self):
        """
        This method should update the current iteration counter, as well as handling any more complicated
        behaviour required for multiple Jones terms. Default version just bumps the iteration counter.
        
        Returns:
            Tuple of two values
                - value of iteration counter
                - True/False hint indicating if a "major" step was taken
        """
        self._iters += 1
        return self.iters, False

    @abstractmethod
    def restrict_solution(self, gains):
        """
        This method should perform any necessary restrictions on the solution, eg. selecting a 
        reference antenna or taking only the amplitude. Function signature must be consistent with 
        the one defined here.
        """

        return NotImplementedError

    @abstractproperty
    def has_converged(self):
        """
        This property must return the convergence status of the gain machine. 
        """
        return NotImplementedError

    @abstractproperty
    def has_stalled(self):
        """
        This property must return the convergence stall status of the gain machine. Note that it
        may be assigned to in the solver.
        """
        return NotImplementedError

    @abstractproperty
    def conditioning_status_string(self):
        """
        This property must return a conditioning status string for the gain machine, e.g.
        
        """
        return NotImplementedError

    @abstractproperty
    def current_convergence_status_string(self):
        """
        This property must return a convergence status string for the gain machine, e.g.
            "G: 20 iters, conv 60.02%, g/fl 15.00%, max update 1e-3"
        This is called while iterating.
        """
        return NotImplementedError

    @property
    def final_convergence_status_string(self):
        """
        This property must return a convergence status string for the gain machine, e.g.
            "G: 20 iters, conv 60.02%, g/fl 15.00%, max update 1e-3"
        This is called once the solution is finished, so it may report a slightly different
        status. Default version uses the current status.
        """
        return self.current_convergence_status_string


    @staticmethod
    def exportable_solutions():
        """
        Returns dict of {label: (empty_value, axes_list)} describing the types of parameters that
        this machine can export. Axes is a list of axis labels. Static method, as it is called
        before any GM is actually created. If empty_value is float or complex, global precision 
        settings will be used.
        """

        return {}

    def importable_solutions(self, grid0):
        """
        Returns dict of parameters that this machine can import, as {label: grid_dict}. Grid_dict 
        knows which grid the parameters must be interpolated onto for this machine. Called when a 
        machine has been created (so grids are available).
        """

        return {}

    @abstractmethod
    def export_solutions(self):
        """
        This method returns the solutions as a dict of {label: masked_array, grid} elements. Arrays
        are masked since solutions have associated flags. Labels must be consistent with whatever 
        exportable_solutions() returns, but not all solutions promised by exportable_solutions() 
        will actually need to be exported.
        
        Grid is a dict, defining axes on which solutions are given, e.g. {'time': vector, 
        'freq': vector}. Note that axes which are fully spanned (e.g. antenna, correlation) need 
        not be present in the grid.
        """

        return NotImplementedError

    @abstractmethod
    def import_solutions(self, solutions_dict):
        """
        This method loads the internal solutions from a dict of {label: array} elements. Labels 
        must be in importable_solutions. Array are masked since solutions have associated flags.
        Array shapes will conform to importable_solutions() results.
        """

        return NotImplementedError


    @classmethod
    def create_factory(machine_cls, *args, **kw):
        """
        This method creates a Machine Factory that will initialize the correct type of gain machine.
        This must be called via the subclasses, so that the factory gets the proper class 
        information.
        """

        return machine_cls.Factory(machine_cls, *args, **kw)

    def _load_solutions(self, init_sols, full_grid):
        """
        Helper method invoked by Factory.create_machine() to import existing solutions into machine.
        Looks for solutions corresponding to this machine's jones_label in init_sols, and
        invokes import_solutions() as appropriate.
        
        Args:
            init_sols: dict of initial solutions, given as {jones_label:(database, prefix)}
        """
        sols = {}
        # collect importable solutions from DB, interpolate
        for label, grids in self.importable_solutions(full_grid).items():
            db, prefix, interpolate = init_sols.get(self.jones_label, (None, None, False))
            name = "{}:{}".format(prefix, label)
            if db is not None:
                if name in db:
                    # check for matching grids
                    mismatches = db[name].find_mismatched_grids(**grids)
                    if mismatches:
                        # are non-interpolatable axes also missing?
                        missing_solutions = not all([interpolatable for _,_,interpolatable in mismatches])
                        # info if only interpolatable entries missing, else warning
                        if interpolate and not missing_solutions:
                            report_func = log(0).print
                        else:
                            report_func = log.warn
                        report_func("{}: {} grid not exactly matched in {}, interpolation may be required".format(
                                self.chunk_label, name, db.filename))
                        # report on axes one by one
                        for axis, values, interpolatable in mismatches:
                            report_func = log(0).print if interpolate and interpolatable else log.warn
                            report_func("    {}: {} not matched{}".format(axis,
                                                     "{} entries".format(len(values)) if len(values) > 5 else
                                                     ", ".join(map(str, values)),
                                                    (", will interpolate" if interpolate and interpolatable else "")))
                    # if interpolating, this is allowed -- otherwise crash out
                    if interpolate:
                        print("{}: interpolating {} from {}".format(self.chunk_label, name, db.filename), file=log)
                        sols[label] = sol = db[name].reinterpolate(**grids)
#                        import ipdb; ipdb.set_trace()
                    else:
                        if mismatches:
                            raise ValueError("{} does not define {} on the correct grid. Consider using "
                                             "-xfer-from rather than -load-from".format(name, db.filename))
                        print("{}: loading {} from {}".format(self.chunk_label, name, db.filename), file=log)
                        sols[label] = sol = db[name].lookup(**grids)
                    if sol.count() != sol.size:
                        print("{}: {:.2%} valid {} slots populated".format(
                            self.chunk_label, sol.count()/float(sol.size), name), file=log)
                    db[name].release_cache()
                else:
                    print("{}: {} not in {}".format(self.chunk_label, name, db.filename), file=log)
        # if anything at all was loaded from DB, import
        if sols:
            self.import_solutions(sols)

    @classmethod
    def raise_userwarning(self, level, msg, sort_index, verbosity=0, raise_once=None, color=None):
        """
        Raise user warning to be summarized after solving
        """
        MasterMachine.__user_warnings.append({
            "level": level,
            "msg": msg,
            "sort_index": sort_index,
            "raise_once": raise_once,
            "verbosity": verbosity,
            "color": color})

    @classmethod
    def collect_warnings(self):
        """
        Sorted list of user warnings raised during solving
        """
        res = sorted(MasterMachine.__user_warnings, key=lambda d: d.get("sort_index", 0))
        MasterMachine.__user_warnings = []
        return res

    class Factory(object):
        """
        A Machine Factory class is created as a singleton: it is responsible for creating and initializing 
        individual gain machines for chunks of the solution space. The purpose of a factory is to hold "global" 
        information for the solution overall (e.g. the overall time/frequency range etc.)
        
        TODO: directions must also be passed in
        """

        def __init__(self, machine_cls, grid, double_precision, apply_only, global_options, jones_options):
            """
            Initializes a factory

            Args:
                machine_cls: the class of the gain machine that will be created
                
                grid: the grid on which solutions are expected, for axes that are already known (antennas, correlations)
                  A dictionary of {axis_name:values_vector}. time/freq need not be supplied, as it is filled in chunk by chunk.
                  
                double_precision: if True, gain machines will use 64-bit float arithmetic
                
                apply_only: if True, solutions are only applied, not solved for
                
                global_options: dict of global options for CubiCal (GD)
                
                jones_options: dict of Jones term options for the specific Jones term
            """
            self.global_options = global_options
            self.jones_options = jones_options
            self.jones_label = self.jones_options['label']
            self.machine_class = machine_cls
            if apply_only:
                self.jones_options['solvable'] = False
            self.solvable = self.jones_options['solvable']
            self.grid = grid
            self.ctype = np.complex128 if double_precision else np.complex64
            self.ftype = np.float64    if double_precision else np.float32
            # dict of jones label -> param database object to init from
            self._init_sols = {}
            # dict of jones label -> param database object to save to. Multiple jones labels
            # may share the same DB (_save_sols_byname holds a unique dict)
            self._save_sols = {}
            # dict of filename -> unique param database object.
            self._save_sols_byname = {}
            # initialize solution databases
            self.init_solutions()

        def get_allocators(self):
            """
            Returns allocation functions appropriate for the class of the gain machine.
            Returns tuple of vis_allocator, flag_allocator.
            """
            return self.machine_class.get_allocators(self.jones_options)

        def init_solutions(self):
            """
            Internal method, called from constructor. Initializes solution databases.
            Default behaviour uses the _init_solutions() implementation below.
            Note that this is reimplemented in JonesChain to collect solution info from chain.
            """
            self._init_solutions(self.jones_label,
                                 self.make_filename(self.jones_options["xfer-from"]) or
                                 self.make_filename(self.jones_options["load-from"]),
                                 bool(self.jones_options["xfer-from"]),
                                 self.solvable and self.make_filename(self.jones_options["save-to"]),
                                 self.machine_class.exportable_solutions(),
                                 dd_term=bool(self.jones_options["dd-term"]))

        def make_filename(self, filename, jones_label=None):
            """
            Helper method: expands full filename a from templated filename. This uses the standard 
            str.format() function, passing in self.global_options, as well as JONES=jones_label, as keyword 
            arguments. This allows for filename templates that include strings from the global options
            dictionary, e.g. "{data[ms]}-ddid{sel[ddid]}".
            
            Args:
                filename (str): 
                    the templated filename
                jones_label (str, optional):
                    Jones matrix label, overrides self.jones_label if specified.
                
            Returns:
                str:
                    Expanded filename
                
            """
            from cubical.main import expand_templated_name
            return expand_templated_name(filename,
                                         JONES=jones_label or self.jones_label)

        def _init_solutions(self, label, load_from, interpolate, save_to, exportables, dd_term=False):
            """
            Internal helper implementation for init_solutions(): this initializes a pair of solution databases.
            
            Args:
                label (str): 
                    the Jones matrix label
                load_from (str):
                    filename of solution DB to load from. Can be empty or None.
                interpolate (bool):
                    True if solutions are allowed to be interpolated.
                save_to (str):
                    filename of solution DB to save to. Can be empty or None.
                exportables (dict):
                    Dictionary of {key: (empty_value, axis_list)} describing the solutions that will be saved.
                    As returned by exportable_solutions() of the gain machine.
                dd_term (bool):
                    True if paremeter was configured as direction-dependent
                
            """
            # init solutions from database
            if load_from:
                print("{} solutions will be initialized from {}".format(label, load_from), file=log(0, "blue"))
                if "//" in load_from:
                    filename, prefix = load_from.rsplit("//", 1)
                else:
                    filename, prefix = load_from, label
                self._init_sols[label] = param_db.load(filename), prefix, interpolate
            # create database to save to
            if save_to:
                # define parameters in DB
                for sol_label, (empty_value, axes) in exportables.items():
                    self.define_param(save_to, "{}:{}".format(label, sol_label), empty_value, axes, dd_term=dd_term)
                print("{} solutions will be saved to {}".format(label, save_to), file=log(0))

        def define_param(self, save_to, name, empty_value, axes,
                          interpolation_axes=("time", "freq"), dd_term=False):
            """
            Internal helper method for _init_solutions(). Defines a parameter to be saved.
            
            Args:
                save_to (str):
                    filename of solution DB to save to. Can be empty or None.
                name (str):
                    name of parameter to be saved
                empty_value:
                    Scalar representing an empty (default) solution, e.g. zero or unity of int, float or complex type.
                axes (iterable):
                    List of axes over which the parameter is defined, e.g. ["time", "freq", "ant1", "ant2"]
                interpolation_axes (iterable):
                    List of axes over which the parameter can be interpolated. Subset of axes.
                dd_term (bool):
                    True if paremeter was configured as direction-dependent
            """
            # init DB, if needed
            db = self._save_sols_byname.get(save_to)
            if db is None:
                selfield = self.global_options["sel"]["field"]
                assert type(selfield) is int, "Currently only supports single field data selection"
                selddid = self.global_options["sel"]["ddid"]
                #assert ((type(selddid) is list) and (all([type(t) is int for t in selddid]))) or \
                #       (type(selddid) is int) or selddid is None, "SPW should be a list of ints or int or None. This is a bug"
                meta = {"field": selfield}
                self._save_sols_byname[save_to] = db = param_db.create(save_to, metadata=meta, backup=True)
            self._save_sols[name] = db
            # work out type of empty value
            if type(empty_value) is float:
                dtype = self.ftype
            elif type(empty_value) is complex:
                dtype = self.ctype
            else:
                dtype = type(empty_value)
            # build parameter grid from predefined grid. Trailing digits are handled:
            # solution axes such as "ant", "ant1" and "ant2" will map to predefined axis "ant",
            # if this is available.
            grid = {}
            for axis in axes:
                if axis in self.grid:
                    # use saved grid, or else [0] for direction axis of non-DD parameter
                    grid[axis] = self.grid[axis] if axis != "dir" or dd_term else [0]
                elif axis[-1].isdigit() and axis[:-1] in self.grid:
                    grid[axis] = self.grid[axis[:-1]]
            return db.define_param(name, dtype, axes, empty=empty_value,
                                   interpolation_axes=interpolation_axes, grid=grid)

        def export_solutions(self, gm, subdict):
            """
            Exports a slice of solutions from a gain machine into a shared dictionary.
            We use shared memory (shared_dict.SharedDict) objects to pass solutions between
            worker (solver) processes and the I/O process which ultimately saves them. This 
            method is called on the solver process side to populate the shared dict.
            
            Args:
                gm:
                    An instance of a gain machine
                    
                subdict (shared_dict.SharedDict):
                    Shared dictionary to be populated (this is generally a subdict of some
                    larger shared dictionary, uniquely assigned to this solver process)
            """
            if not self.solvable:
                return
            sols = gm.export_solutions()
            # has the gain machine added a prefix to the names already (as the chain machine does)
            is_prefixed = sols.pop('prefixed', False)
            # populate values subdictionary
            for label, (value, grid) in sols.items():
                name = label if is_prefixed else "{}:{}".format(gm.jones_label, label)
                subdict[name] = value.data
                subdict["{}:grid__".format(name)]  = grid
                subdict["{}:flags__".format(name)] = value.mask

        def get_solution_db(self, name):
            """
            Returns (output) solution database corresponding to the named parameter
            """
            return self._save_sols[name]

        def save_solutions(self, subdict):
            """
            Saves a slice of the solutions from a dictionary to the database. This is called in the I/O process
            to actually save the solutions that are exported by export_solutions().

            Args:
                subdict (shared_dict.SharedDict):
                    Shared dictionary to be saved. This is presumed to be populated by export_solutions() above.
            """
            # add slices for all parameters
            for name in subdict.keys():
                if not name.endswith("__") and name in self._save_sols:
                    sd = subdict["{}:grid__".format(name)]
                    grids = {key: sd[key] for key in sd.keys()}
                    log(1).print("saving solutions for {}".format(name))
                    self.get_solution_db(name).add_chunk(name, masked_array(subdict[name],
                                                                       subdict[name+":flags__"]), grids)
        def close(self):
            """
            Closes all solution databases and releases various caches.
            """
            for db, prefix, _ in list(self._init_sols.values()):
                db.close()
            for db in list(self._save_sols_byname.values()):
                db.close()
            self._init_sols = {}
            self._save_sols = {}
            self._save_sols_byname = {}

        def create_machine(self, data_arr, n_dir, n_mod, times, freqs, chunk_label):
            """
            Creates a gain machine, given a model, and a time and frequency subset of the global solution space.

            Args:
                data_arr: model array, of shape (nmod,ntim,nfreq,nant,nant,ncorr,ncorr)
                times: times
                freqs: frequencies

            Returns:
                An instance of a gain machine
            """
            gm = self.machine_class(self.jones_label, data_arr, n_dir, n_mod, times, freqs,
                                    chunk_label, self.jones_options)
            gm._load_solutions(self._init_sols, self.grid)
            return gm
        
        def set_metadata(self, src):
            """
            Sets database metadata source
            
            Args:
                src: instance of cubical.data_handler
            """
            self.metadata = src.metadata
            for db in list(self._save_sols_byname.values()):
                db.export_CASA_gaintable = self.global_options["out"].get("casa-gaintables", True)
                db.set_metadata(src)

        def determine_allocators(self):
            return self.machine_class.determine_allocators(self.jones_options)



# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details
"""
Handles parameter databases which can contain solutions and other relevant values. 
"""
from __future__ import print_function
from future.moves import pickle
import os, os.path
import numpy as np
import traceback
from cubical.tools import logger, ModColor
log = logger.getLogger("param_db")
import time
from collections import OrderedDict, Iterator

from cubical.database.parameter import Parameter, _Record
from .iface_database import iface_database
class _ParmSegment(_Record):
    """
    A ParmSegment is just a Record -- we just want it to be a special type so that it
    can be identified when unpickled.
    """
    pass

class PickledDatabase(iface_database):
    """
    This class implements a simple parameter database saved to a pickle.
    """

    def __init__(self):
        self._fobj = None

    def _create(self, filename, metadata={}, backup=True, **kw):
        """
        Creates a parameter database given by the filename and opens it in "create" mode.

        Args:
            filename (str): 
                Name of database.
            metadata (dict, optional): 
                Optional metadata to be stored in DB.
            backup (bool, optional):
                If True, and an old database with the same filename exists, make a backup.
            kw (dict):
                Keyword arguments.
        """

        self.mode = "create"
        self.filename = filename
        self.do_backup = backup
        self.metadata = OrderedDict(mode=self.MODE_FRAGMENTED, time=time.time(), **metadata)
        # we'll write to a temp file, and do a backup on successful closure
        self._fobj = open(filename + ".tmp", 'wb')
        pickle.dump(self.metadata, self._fobj)
        self._fobj.flush()
        self._parameters = {}
        self._parm_written = set()
        print("creating {} in {} mode".format(self.filename, self.metadata['mode']), file=log(0))

    def define_param(self, *args, **kw):
        """
        Defines a parameter. Only valid in "create" mode.

        Args:
            args (tuple):
                Positional arguments.
            kw (dict):
                Keyword arguments.
        """

        assert (self.mode is "create")
        parm = Parameter(*args, **kw)
        self._parameters[parm.name] = parm
        # we don't write it to DB yet -- write it in add_chunk() rather
        # this makes it easier to deal with I/O workers (all IO is done by one process)
        return parm

    def add_chunk(self, name, array, grid={}):
        """
        Adds a slice of values for a parameter.

        Args:
            name (str):
                The parameter name e.g. "G".
            array (:obj:`~numpy.ma.core.MaskedArray`):
                The values which are to be added.
            grid (dict, optional):
                Grid coordinates for each sliced parameter axis. 

        """

        assert (self.mode is "create")
        parm = self._parameters.get(name)
        assert (parm is not None)
        # dump parm to DB the first time a slice shows up
        if name not in self._parm_written:
            pickle.dump(parm, self._fobj, 2)
            self._parm_written.add(name)
        # update axis shapes and grids based on slice
        parm._update_shape(array.shape, grid)
        # dump slice to DB
        item = _ParmSegment(name=name, array=np.ma.asarray(array), grid=grid)
        pickle.dump(item, self._fobj, 2)

    def close(self):
        """ Closes the database. """

        if self._fobj:
            self._fobj.close()
            self._fobj = None
        # in create mode, update the descriptions file
        if self.mode is "create":
            self._save_desc()
            self._backup_and_rename(self.do_backup)
        self.mode = "closed"

    def _save_desc(self):
        """ Helper function. Writes accumulated parameter descriptions to filename.desc. """

        for desc in self._parameters.values():
            desc._finalize_shape()
        for key in list(self._parameters.keys()):
            if not self._parameters[key]._populated:
                del self._parameters[key]
        with open(self.filename + ".skel", 'wb') as pf:
            pickle.dump(self._parameters, pf, 2)
        print("saved updated parameter skeletons to {}".format(self.filename + ".skel"), file=log(0))

    def _backup_and_rename(self, backup):
        """
        May create a backup and renames the temporary DB.

        Args:
            backup (bool):
                If True, creates a backup, otherwise just renames.
        """
        if os.path.exists(self.filename):
            if backup:
                backup_filename = os.path.join(os.path.dirname(self.filename),
                                               "~" + os.path.basename(self.filename))
                print("previous DB will be backed up as " + backup_filename, file=log(0))
                if os.path.exists(backup_filename):
                    print("  removing old backup " + backup_filename, file=log(0))
                    os.unlink(backup_filename)
                os.rename(self.filename, backup_filename)
            else:
                os.unlink(self.filename)
        os.rename(self.filename + ".tmp", self.filename)
        print("wrote {} in {} mode".format(self.filename, self.metadata['mode']), file=log(0))

    def save(self, filename=None, backup=True):
        """
        Save the database.

        Args:
            filename (str, optional):
                Name of output file.
            backup (bool, optional):
                If True, create a backup.
        """
        assert (self.mode is "load")
        self.metadata['mode'] = self.MODE_CONSOLIDATED
        filename = filename or self.filename
        with open(filename + ".tmp", 'w') as fobj:
            pickle.dump(self.metadata, fobj, 2)
            for parm in self._parameters.values():
                parm.release_cache()
            pickle.dump(self._parameters, fobj, 2)
        # successfully written? Backup and rename
        self.filename = filename
        self._backup_and_rename(backup)

    MODE_FRAGMENTED = "fragmented"
    MODE_CONSOLIDATED = "consolidated"

    class _Unpickler(Iterator):
        def __init__(self, filename):
            self.fobj = open(filename, 'rb')
            self.metadata = pickle.load(self.fobj)
            if type(self.metadata) is not OrderedDict or not "mode" in self.metadata:
                raise IOError("{}: invalid metadata entry".format(filename))
            self.mode = self.metadata['mode']
        
        def __del__(self):
            self.fobj.close()

        def __next__(self):
            try:
                return pickle.load(self.fobj)
            except EOFError:
                raise StopIteration

        def next(self):
            return self.__next__()

    def _load(self, filename):
        """
        Loads database from file. This will create arrays corresponding to the stored parameter
        shapes.

        Args:
            filename (str):
                Name of file to load.
        """

        self.mode = "load"
        self.filename = filename

        db = self._Unpickler(filename)
        print("reading {} in {} mode".format(self.filename, db.mode), file=log(0))
        self.metadata = db.metadata
        for key, value in self.metadata.items():
            if key != "mode":
                print("  metadata '{}': {}".format(key, value), file=log(1))

        # now load differently depending on mode
        # in consolidated mode, just unpickle the parameter objects
        if db.mode == PickledDatabase.MODE_CONSOLIDATED:
            self._parameters = next(db)
            for parm in self._parameters.values():
                print("  read {} of shape {}".format(parm.name,
                                                               'x'.join(map(str, parm.shape))), file=log(1))
            return

        # otherwise we're in fragmented mode
        if db.mode != PickledDatabase.MODE_FRAGMENTED:
            raise IOError("{}: invalid mode".format(self.filename, self.metadata.mode))

        # in fragmented mode? try to read the desc file
        descfile = filename + '.skel'
        self._parameters = None
        if not os.path.exists(descfile):
            print(ModColor.Str("{} does not exist, will try to rebuild".format(descfile)), file=log(0))
        elif os.path.getmtime(descfile) < os.path.getmtime(self.filename):
            print(ModColor.Str("{} older than database: will try to rebuild".format(descfile)), file=log(0))
        elif os.path.getmtime(descfile) < os.path.getmtime(__file__):
            print(ModColor.Str("{} older than this code: will try to rebuild".format(descfile)), file=log(0))
        else:
            try:
                with open(descfile, 'rb') as pf:
                    self._parameters = pickle.load(pf)
            except:
                traceback.print_exc()
                print(ModColor.Str("error loading {}, will try to rebuild".format(descfile)), file=log(0))
        # rebuild the skeletons, if they weren't loaded
        if self._parameters is None:
            self._parameters = {}
            for item in db:
                if isinstance(item, Parameter):
                    self._parameters[item.name] = item
                elif isinstance(item, _ParmSegment):
                    self._parameters[item.name]._update_shape(item.array.shape, item.grid)
                else:
                    raise IOError("{}: unexpected entry of type '{}'".format(self.filename, type(item)))
            self._save_desc()

        # initialize arrays
        for parm in self._parameters.values():
            parm._init_arrays()

        # go over all slices to paste them into the arrays
        db = self._Unpickler(filename)
        for item in db:
            if type(item) is Parameter:
                pass
            elif type(item) is _ParmSegment:
                parm = self._parameters.get(item.name)
                if parm is None:
                    raise IOError("{}: no parm found for {}'".format(filename, item.name))
                parm._paste_slice(item)
            else:
                raise IOError("{}: unknown item type '{}'".format(filename, type(item)))

        # ok, now arrays and flags each contain a full-sized array. Break it up into slices.
        for parm in self._parameters.values():
            parm._finalize_arrays()

    def names(self):
        """ Returns names of all defined parameters. """

        return list(self._parameters.keys())

    def __contains__(self, name):
        return name in self._parameters

    def __getitem__(self, name):
        return self.get(name)

    def get(self, name):
        """
        Gets Parameter object associated with the named parameter.

        Args:
            name (str):
                Name of parameter.

        Returns:
            :obj:`~cubical.param_db.Parameter`:
                The requested Parameter object.
        """

        return self._parameters[name]


# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details
"""
Implements the solver loop.
"""
from __future__ import print_function
import numpy as np
import os, os.path
import gc
import traceback
from cubical.tools import logger, ModColor
from cubical.flagging import FL
from cubical.statistics import SolverStats
from cubical.tools import BREAK  # useful: can set static breakpoints by putting BREAK() in the code

## uncomment this to make UserWarnings (from e.g. numpy.ma) into full-blown exceptions
## TODO: add a --debug-catch-warnings option for this?
#import warnings
#warnings.simplefilter('error', UserWarning)
#warnings.simplefilter('error', RuntimeWarning)

from .madmax.flagger import Flagger

log = logger.getLogger("solver")
#log.verbosity(2)

# global defaults dict
GD = None 

# MS metadata
metadata = None

# gain machine factory to use
gm_factory = None

# IFR-based gain machine to use
ifrgain_machine = None

# set to true for old-style (version <= 1.2.1) weight averaging, where 2x2 weights are collapsed into a single number
legacy_version12_weights = False

import builtins
try:
    builtins.profile
except AttributeError:
    # No line profiler, provide a pass-through version
    def profile(func): return func
    builtins.profile = profile

@builtins.profile
def _solve_gains(gm, stats, madmax, obser_arr, model_arr, flags_arr, sol_opts, label="", compute_residuals=None):
    """
    Main body of the GN/LM method. Handles iterations and convergence tests.

    Args:
        gm (:obj:`~cubical.machines.abstract_machine.MasterMachine`): 
            The gain machine which will be used in the solver loop.
        obser_arr (np.ndarray): 
            Shape (n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing observed 
            visibilities. 
        model_arr (np.ndarray): 
            Shape (n_dir, n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing model 
            visibilities. 
        flags_arr (np.ndarray): 
            Shape (n_tim, n_fre, n_ant, n_ant) integer array containing flag data.
        sol_opts (dict): 
            Solver options (see [sol] section in DefaultParset.cfg).
        label (str, optional):             
            Label identifying the current chunk (e.g. "D0T1F2").
        compute_residuals (bool, optional): 
            If set, the final residuals will be computed and returned.

    Returns:
        2-element tuple
            
            - resid (np.ndarray)
                The final residuals (if compute_residuals is set), else None.
            - stats (:obj:`~cubical.statistics.SolverStats`)
                An object containing solver statistics.
    """
    chi_interval = sol_opts["chi-int"]
    stall_quorum = sol_opts["stall-quorum"] 

    diverging = ""

    # for all the solvers that do not output any weights and for the robust solver when they are no valid solutions
    gm.output_weights = None 

    # Estimates the overall noise level and the inverse variance per channel and per antenna as
    # noise varies across the band. This is used to normalize chi^2.

    stats.chunk.noise, inv_var_antchan, inv_var_ant, inv_var_chan = \
                                        stats.estimate_noise(obser_arr, flags_arr)

    # if we have directions in the model, but the gain machine is non-DD, collapse them
    if not gm.dd_term and model_arr.shape[0] > 1:
        model_arr = model_arr.sum(axis=0, keepdims=True)

    # This works out the conditioning of the solution, sets up various chi-sq normalization
    # factors etc, and does any other precomputation required by the current gain machine.

    gm.precompute_attributes(obser_arr, model_arr, flags_arr, inv_var_chan)

    # apply any flags raised in the precompute

    new_flags = flags_arr & ~(FL.MISSING | FL.PRIOR) != 0
    model_arr[:, :, new_flags, :, :] = 0
    obser_arr[:, new_flags, :, :] = 0

    def get_flagging_stats():
        """Returns a string describing per-flagset statistics"""
        fstats = []

        for flag, mask in FL.categories().items():
            n_flag = ((flags_arr & mask) != 0).sum()
            if n_flag:
                fstats.append("{}:{}({:.2%})".format(flag, n_flag, n_flag/float(flags_arr.size)))

        return " ".join(fstats)

    def update_stats(flags, statfields):
        """
        This function updates the solver stats object with a count of valid data points used for chi-sq 
        calculations
        """
        unflagged = (flags==0)
        # Compute number of terms in each chi-square sum. Shape is (n_tim, n_fre, n_ant).

        nterms  = 2 * gm.n_cor * gm.n_cor * np.sum(unflagged, axis=3)

        # Update stats object accordingly.

        for field in statfields:
            getattr(stats.chanant,  field+'n')[...] = np.sum(nterms, axis=0)
            getattr(stats.timeant,  field+'n')[...] = np.sum(nterms, axis=1)
            getattr(stats.timechan, field+'n')[...] = np.sum(nterms, axis=2)
    
    update_stats(flags_arr, ('initchi2', 'chi2'))

    # Initialize a residual array.

    resid_shape = [gm.n_mod, gm.n_tim, gm.n_fre, gm.n_ant, gm.n_ant, gm.n_cor, gm.n_cor]

    resid_arr = gm.allocate_vis_array(resid_shape, obser_arr.dtype, zeros=True)
    gm.compute_residual(obser_arr, model_arr, resid_arr, require_full=True)
    resid_arr[:,flags_arr!=0] = 0

    # apply MAD flagging
    madmax.set_mode(GD['madmax']['enable'])

    # do mad max flagging, if requested
    thr1, thr2 = madmax.get_mad_thresholds()
    if thr1 or thr2:
        if madmax.beyond_thunderdome(resid_arr, obser_arr, model_arr, flags_arr, thr1, thr2, "{} initial".format(label)):
            gm.update_equation_counts(flags_arr != 0)
            stats.chunk.num_mad_flagged = ((flags_arr & FL.MAD) != 0).sum()


    # apply robust flag if robust machine (this uses the madmax flag)
    if hasattr(gm, 'is_robust'):
        if gm.robust_flag_weights:
            gm.robust_flag(flags_arr, model_arr, obser_arr)
            stats.chunk.num_mad_flagged = ((flags_arr & FL.MAD) != 0).sum()

    # In the event that there are no solutions with valid data, this will log some of the
    # flag information and break out of the function.
    stats.chunk.num_solutions = gm.num_solutions
    stats.chunk.num_sol_flagged = gm.num_gain_flags()[0]

    # every chunk stat set above now copied to stats.chunk.field_0
    stats.save_chunk_stats(step=0)

    # raise warnings from priori conditioning, before the loop
    for d in gm.collect_warnings():
        log.write(d["msg"],
                  level=d["level"],
                  print_once=d["raise_once"],
                  verbosity=d["verbosity"],
                  color=d["color"])

    if not gm.has_valid_solutions:
        log.error("{} no solutions: {}; flags {}".format(label, gm.conditioning_status_string, get_flagging_stats()))
        return (obser_arr if compute_residuals else None), stats, None

    def compute_chisq(statfield=None, full=True):
        """
        Computes chi-squared statistic based on current residuals and noise estimates.
        Populates the stats object with it.

        Full=True at the beginning and end of a solution, and it is passed to gm.compute_chisq()
        """
        chisq, chisq_per_tf_slot, chisq_tot = gm.compute_chisq(resid_arr, inv_var_chan, require_full=full)

        if statfield:
            getattr(stats.chanant, statfield)[...]  = np.sum(chisq, axis=0)
            getattr(stats.timeant, statfield)[...]  = np.sum(chisq, axis=1)
            getattr(stats.timechan, statfield)[...] = np.sum(chisq, axis=2)

        return chisq_per_tf_slot, chisq_tot

    chi, stats.chunk.chi2u = compute_chisq(statfield='initchi2', full=True)
    stats.chunk.chi2_0 = stats.chunk.chi2u_0 = stats.chunk.chi2u

    # The following provides conditioning information when verbose is set to > 0.
    if log.verbosity() > 0:
        log(1).print("{} chi^2_0 {:.4}; {}; noise {:.3}, flags: {}".format(
                        label, stats.chunk.chi2_0, gm.conditioning_status_string,
                        float(stats.chunk.noise_0), get_flagging_stats()))

    # Main loop of the NNLS method. Terminates after quorum is reached in either converged or
    # stalled solutions or when the maximum number of iterations is exceeded.

    
    major_step = 0  # keeps track of "major" solution steps, for purposes of collecting stats


    while not(gm.has_converged) and not(gm.has_stalled):

        num_iter, update_major_step = gm.next_iteration()

        if update_major_step:
            major_step += 1
            stats.save_chunk_stats(step=major_step)

        # This is currently an awkward necessity - if we have a chain of jones terms, we need to
        # make sure that the active term is correct and need to support some sort of decision making
        # for testing convergence. I think doing the iter increment here might be the best choice,
        # with an additional bit of functionality for Jones chains. I suspect I will still need to
        # change the while loop component to be compatible with the idea of partial convergence.
        # Perhaps this should all be done right at the top of the function? A better idea is to let
        # individual machines be aware of their own stalled/converged status, and make those
        # properties more complicated on the chain. This should allow for fairly easy substitution
        # between the various machines.

        gm.compute_update(model_arr, obser_arr)

        # flag solutions. This returns True if any flags have been propagated out to the data.
        if gm.flag_solutions(flags_arr, False):

            update_stats(flags_arr, ('chi2',))

            # Re-zero the model and data at newly flagged points.
            # TODO: is this needed?
            # TODO: should we perhaps just zero the model per flagged direction, and only flag the data?
            # OMS: probably not: flag propagation is now handled inside the gain machine. If a flag is
            # propagated out to the data, then that slot is gone gone gone and should be zeroe'd everywhere.

            new_flags = flags_arr&~(FL.MISSING|FL.PRIOR) !=0
           
            model_arr[:, :, new_flags, :, :] = 0
            obser_arr[   :, new_flags, :, :] = 0

            stats.chunk.num_sol_flagged = gm.num_gain_flags()[0]


        # Compute values used in convergence tests. This check implicitly marks flagged gains as
        # converged.

        gm.check_convergence(gm.epsilon)

        stats.chunk.iters = num_iter
        stats.chunk.num_converged = gm.num_converged_solutions
        stats.chunk.frac_converged = gm.num_solutions and gm.num_converged_solutions / float(gm.num_solutions)

        # Break out of the solver loop if we find ourselves with no valid solution intervals (e.g. due to gain flagging)
        if not gm.has_valid_solutions:
            break

        # Check residual behaviour after a number of iterations equal to chi_interval. This is
        # expensive, so we do it as infrequently as possible.

        if (num_iter % chi_interval) == 0 or num_iter <= 1:

            old_chi, old_mean_chi = chi, float(stats.chunk.chi2u)

            gm.compute_residual(obser_arr, model_arr, resid_arr, require_full=False)
            resid_arr[:,flags_arr!=0] = 0

            # do mad max flagging, if requested
            thr1, thr2 = madmax.get_mad_thresholds()
            if thr1 or thr2:
                num_mad_flagged_prior = int(stats.chunk.num_mad_flagged)

                if madmax.beyond_thunderdome(resid_arr, obser_arr, model_arr, flags_arr, thr1, thr2,
                                             "{} iter {} ({})".format(label, num_iter, gm.jones_label)):
                    gm.update_equation_counts(flags_arr != 0)
                    stats.chunk.num_mad_flagged = ((flags_arr&FL.MAD) != 0).sum()
                    if stats.chunk.num_mad_flagged != num_mad_flagged_prior:
                        log(2).print("{}: {} new MadMax flags".format(label,
                                        stats.chunk.num_mad_flagged - num_mad_flagged_prior))

            chi, stats.chunk.chi2u = compute_chisq(full=False)

            # Check for stalled solutions - solutions for which the residual is no longer improving.
            # Don't do this on a major step (i.e. when going from term to term in a chain), as the
            # reduced chisq (which compute_chisq() returns) can actually jump when going to the next term

            if update_major_step:
                stats.chunk.num_stalled = stats.chunk.num_diverged = 0
            else:
                delta_chi = old_chi - chi
                stats.chunk.num_stalled = np.sum((delta_chi <= gm.delta_chi*old_chi))
                diverged_tf_slots = delta_chi < -0.1 * old_chi
                stats.chunk.num_diverged = diverged_tf_slots.sum()
                # at first iteration, flag immediate divergers
                if sol_opts['flag-divergence'] and stats.chunk.num_diverged and num_iter == 1:
                    model_arr[:, :, diverged_tf_slots] = 0
                    obser_arr[:, diverged_tf_slots] = 0

                    # find previously unflagged visibilities that have become flagged due to divergence
                    new_flags = (flags_arr == 0)
                    new_flags[~diverged_tf_slots] = 0

                    flags_arr[diverged_tf_slots] |= FL.DIVERGE

                    num_nf = new_flags.sum()
                    log.warn("{}: {:.2%} slots diverging, {} new data flags".format(label,
                                    diverged_tf_slots.sum()/float(diverged_tf_slots.size), num_nf))

            stats.chunk.frac_stalled = stats.chunk.num_stalled / float(chi.size)
            stats.chunk.frac_diverged = stats.chunk.num_diverged / float(chi.size)

            gm.has_stalled = (stats.chunk.frac_stalled >= stall_quorum)

            # if gm.has_stalled:
            #     import pdb; pdb.set_trace()

            if log.verbosity() > 1:
                if update_major_step:
                    delta_chi_max = delta_chi_mean = 0.
                else:
                    wh = old_chi != 0
                    delta_chi[wh] /= old_chi[wh]
                    delta_chi_max  = delta_chi.max()
                    chi_mean = float(stats.chunk.chi2u)
                    delta_chi_mean = (old_mean_chi - chi_mean) / chi_mean if chi_mean != 0 else 0.

                if stats.chunk.num_diverged:
                    diverging = ", " + ModColor.Str("diverging {:.2%}".format(stats.chunk.frac_diverged), "red")
                else:
                    diverging = ""

                log(2).print("{} {} chi2 {:.4}, rel delta {:.4} max {:.4}, active {:.2%}{}".format(
                                    label, gm.current_convergence_status_string,
                                    stats.chunk.chi2u, delta_chi_mean, delta_chi_max,
                                    float(1-stats.chunk.frac_stalled), diverging))

        # Adding the below lines for the robust solver so that flags should be apply to the weights
        if hasattr(gm, 'is_robust'):
            gm.update_weight_flags(flags_arr)


    # num_valid_solutions will go to 0 if all solution intervals were flagged. If this is not the
    # case, generate residuals etc.

    if gm.has_valid_solutions:
        # Final round of flagging
        flagged = gm.flag_solutions(flags_arr, final=True)
        stats.chunk.num_sol_flagged = gm.num_gain_flags(final=True)[0]
    else:
        flagged = None



    # check this again, because final round of flagging could have killed us
    if gm.has_valid_solutions:
        # Do we need to recompute the final residuals?
        if (sol_opts['last-rites'] or compute_residuals):
            gm.compute_residual(obser_arr, model_arr, resid_arr, require_full=True)
            resid_arr[:,flags_arr!=0] = 0

            # do mad max flagging, if requested
            thr1, thr2 = madmax.get_mad_thresholds()
            if thr1 or thr2:
                if madmax.beyond_thunderdome(resid_arr, obser_arr, model_arr, flags_arr, thr1, thr2,
                                            "{} final".format(label)) and sol_opts['last-rites']:
                    gm.update_equation_counts(flags_arr != 0)

        # Re-estimate the noise using the final residuals, if last rites are needed.

        if sol_opts['last-rites']:
            stats.chunk.noise, inv_var_antchan, inv_var_ant, inv_var_chan = \
                                        stats.estimate_noise(resid_arr, flags_arr, residuals=True)
            chi1, stats.chunk.chi2 = compute_chisq(statfield='chi2', full=True)
        else:
            stats.chunk.chi2 = stats.chunk.chi2u

        message = "{} (end solve) {}, stall {:.2%}{}, chi^2 {:.4} -> {:.4}".format(label, gm.final_convergence_status_string,
                  float(stats.chunk.frac_stalled), diverging, float(stats.chunk.chi2_0), stats.chunk.chi2u)

        should_warn = float(stats.chunk.chi2_0) < float(stats.chunk.chi2u) or diverging
        if sol_opts['last-rites'] and (should_warn or log.verbosity() > 0):
            message = "{} ({:.4}), noise {:.3} -> {:.3}".format(message,
                            float(stats.chunk.chi2), float(stats.chunk.noise_0), float(stats.chunk.noise))
        if should_warn:
            message += " Shows signs of divergence. If you see this message often you may have significant RFI present in your data or your solution intervals are too short."
        if should_warn:
            log.warn(message)
        elif log.verbosity() > 0:
            log.info(message)

    # If everything has been flagged, no valid solutions are generated.

    else:
        log.error("{} (end solve) {}: completely flagged?".format(label, gm.final_convergence_status_string))

        chi2 = chi2u = 0
        resid_arr = obser_arr

    robust_weights = None
    if hasattr(gm, 'is_robust'):
        
        # do a last round of robust flag robust flag and save the weights
         
        if gm.robust_flag_weights and not gm.robust_flag_disable:
            gm.robust_flag(flags_arr, model_arr, obser_arr, final=True)
            stats.chunk.num_mad_flagged = ((flags_arr & FL.MAD) != 0).sum()
        
        if gm.save_weights:
            newshape = gm.weights.shape[1:-1] + (2,2)
            robust_weights = np.repeat(gm.weights.real, 4, axis=-1)
            robust_weights = np.reshape(robust_weights, newshape)
            gm.output_weights = robust_weights
        

    # After the solver loop check for warnings from the solvers
    for d in gm.collect_warnings():
        log.write(d["msg"],
                  level=d["level"],
                  print_once=d["raise_once"],
                  verbosity=d["verbosity"],
                  color=d["color"])


    return (resid_arr if compute_residuals else None), stats, robust_weights


class _VisDataManager(object):
    """A _VisDataManager object holds data, model, flag and weight arrays associated with a single
    chunk of visibility data. It also holds a GainMachine. It provides methods and properties for
    computing/caching various derived arrays (weighted versions of data and model, corrupt models, etc.)
    on demand.

    _VisDataManagers are used to unify the interface to the various solving methods defined below.
    """
    def __init__(self, obser_arr, model_arr, flags_arr, weight_arr, freq_slice):
        """
        Initialises a VisDataManager.

        Args:
            obser_arr (np.ndarray):
                Shape (n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) complex array containing observed visibilities.
            model_arr (np.ndarray):
                Shape (n_dir, n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) complex array containing model
                visibilities.
            flags_arr (np.ndarray):
                Shape (n_tim, n_fre, n_ant, n_ant) integer array containing flags.
            weight_arr (np.ndarray):
                Shape (n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) float array containing weights.
            freq_slice (slice):
                Slice into the full data frequency axis corresponding to this chunk.
        """
        self.gm = None
        ## OMS: take sqrt() of weights since that's the correct thing to use in whitening
        self.obser_arr, self.model_arr, self.flags_arr, self.weight_arr = \
            obser_arr, model_arr, flags_arr, weight_arr
        if GD['data']['normalize']:
            self._normalization = np.abs(obser_arr)
        else:
            self._normalization = None
        if weight_arr is not None:
            if legacy_version12_weights:
                # self.weight_arr[:] = np.sqrt(self.weight_arr.mean(axis=(-1,-2)))[..., np.newaxis, np.newaxis]
                if self._normalization is not None:
                    self.weight_arr = weight_arr*self._normalization
                self.weight_arr[:] = self.weight_arr.mean(axis=(-1,-2))[..., np.newaxis, np.newaxis]
            else:
                np.sqrt(self.weight_arr, out=self.weight_arr)
                if self._normalization is not None:
                    self.weight_arr *= self._normalization
            
        self._wobs_arr = self._wmod_arr = None
        self.freq_slice = freq_slice
        self._model_corrupted = False

    @property
    def weighted_obser(self):
        """
        This property gives the observed visibilities times the weights

        Returns:
            Weighted observed visibilities (np.ndarray) of shape (n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor)
        """
        if self._wobs_arr is None:
            if self.weight_arr is not None:
                self._wobs_arr = self.obser_arr[np.newaxis,...] * self.weight_arr
            else:
                self._wobs_arr = self.obser_arr.copy().reshape([1]+list(self.obser_arr.shape))
                # zero the flagged visibilities. Note that if we have a weight, this is not necessary,
                # as they will already get zero weight in data_handler
                self._wobs_arr[:, self.flags_arr!=0, :, :] = 0
            if self._normalization is not None:
                with np.errstate(divide='ignore', invalid='ignore'):
                    self._wobs_arr /= self._normalization[np.newaxis,...]
                    self._wobs_arr[(self._normalization==0)[np.newaxis,...]] = 0
                self._normalization = None
              
        return self._wobs_arr

    @property
    def weighted_model(self):
        """
        This property gives the model visibilities times the weights

        Returns:
            Weighted model visibilities (np.ndarray) of shape (n_dir, n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor)
        """
        if self._wmod_arr is None:
            if self.weight_arr is not None:
                self._wmod_arr = self.model_arr * self.weight_arr[np.newaxis, ...]
            else:
                self._wmod_arr = self.model_arr.copy()
                # zero the flagged visibilities. Note that if we have a weight, this is not necessary,
                # as they will already get zero weight in data_handler
                self._wmod_arr[:, :, self.flags_arr!=0, :, :] = 0
        return self._wmod_arr

    @property
    def corrupt_weighted_model(self):
        """
        This property gives the model visibilities, corrupted by the gains, times the weights, summed over directions.

        Returns:
            Weighted corrupted model visibilities (np.ndarray) of shape (n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor)
        """
        cmod = self.corrupt_model(None).sum(0)
        if self.weight_arr is not None:
            return cmod*self.weight_arr
        else:
            return cmod

    def corrupt_residual(self, imod=0, idir=slice(None)):
        """
        This method returns the (corrupted) residual with respect to a given model

        Args:
            imod (int):
                Index of model (0 to n_mod-1).

        Returns:
            Weighted residual visibilities (np.ndarray) of shape (n_tim, n_fre, n_ant, n_ant, n_cor, n_cor)
        """
        return self.obser_arr - self.corrupt_model(imod, idir)


    def corrupt_model(self, imod=0, idir=slice(None)):
        """
        This method returns the model visibilities, corrupted by the gains.
        The first time it is called, the model is corrupted in-place.
        Args:
            imod (int or None):
                Index of model (0 to n_mod-1), or None to return all models
            idir (slice or list)
                Directions to include in corrupted moel

        Returns:
            Corrupted model visibilities (np.ndarray) of shape (n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor)
            if imod is None, otherwise the model axis is omitted.
        """
        if not self._model_corrupted:
            self.gm.apply_gains(self.model_arr, full2x2=True)
            self._model_corrupted = True
        if imod is None:
            return self.model_arr
        return self.model_arr[idir, imod].sum(0)


class classproperty(property):
    def __get__(self, cls, owner):
        return classmethod(self.fget).__get__(None, owner)()

class SolverMachine(object):
    """Base class encapsulating different solver methods and their properties"""
    def __init__(self, vdm, gm, soldict, sol_opts, label, metadata):
        """
        Args:
            vdm (:obj:`_VisDataManager`):
                VisDataManager for this chunk of data
            soldict (:obj:~cubical.tools.shared_dict.SharedDict):
                Shared dict used to pass solutions (IFR gain solutions, primarily) back out to the
                calling thread.
            label (str):
                Label identifying the current chunk (e.g. "D0T1F2").
            sol_opts (dict):
                Solver options (see [sol] section in DefaultParset.cfg).
            soldict (:obj:~cubical.tools.shared_dict.SharedDict):
                Shared dict used to pass solutions (IFR gain solutions, primarily) back out to the
                calling thread.
        """
        self.sol_opts = sol_opts
        self.label = label
        self.metadata = metadata
        self.vdm = vdm
        self.gm  = gm
        self.soldict = soldict
        self.stats = SolverStats(vdm.obser_arr)
        self.stats.chunk.label = label
        self.stats.chunk.num_prior_flagged = (
                    vdm.flags_arr & ~(FL.MISSING | FL.SKIPSOL) != 0).sum()  # number of prior flagged data points
        self.stats.chunk.num_data_points = (vdm.flags_arr == 0).sum()  # nominal number of valid data points

        # for apply-only machines, precompute machine attributes and apply initial gain flags
        if self.is_apply_only:
            gm.precompute_attributes(vdm.obser_arr, vdm.model_arr, vdm.flags_arr, None)
            gm.flag_solutions(vdm.flags_arr, True)
            self.stats.chunk.num_solutions = vdm.gm.num_solutions
            self.stats.chunk.num_sol_flagged = vdm.gm.num_gain_flags(final=True)[0]

        # initialize the flagger
        self.madmax = Flagger(GD, label, metadata, self.stats)
        self.madmax.set_mode(GD['madmax']['enable'])

        # output weights, if any, go here
        self.output_weights = None

    # various traits of the solver machine, redefined by subclasses

    # does this machine require model visibilities?
    @classproperty
    def is_model_required(cls):
        return True

    # does this machine produce full corrected residuals?
    @property
    def outputs_full_corrected_residuals(cls):
        return False

    # is this an apply-only (i.e. non-solver) machine?
    is_apply_only = False


    def run(self):
        """
        Abstract method to run the solver.

        Returns:
            np.array of output visibilities
        """
        return NotImplementedError

    def finalize(self, corr_vis):
        """
        Finalizes the output visibilities, running a pass of the flagger on them, if configured
        """
        
        # clear out MAD flags if madmax was in trial mode
        if self.stats.chunk.num_mad_flagged and self.madmax.trial_mode:
            self.vdm.flags_arr &= ~FL.MAD
            self.stats.chunk.num_mad_flagged = 0
        num_mad_flagged_prior = int(self.stats.chunk.num_mad_flagged)

        # apply final round of madmax on residuals, if asked to
        if GD['madmax']['residuals']:
            # recompute the residuals if required
            if self.outputs_full_corrected_residuals:
                resid_vis = corr_vis
                log(0).print("{}: doing final MadMax round on residuals".format(self.label))
            else:
                log(0).print("{}: computing full residuals for final MadMax round".format(self.label))
                resid_vis1 = self.vdm.corrupt_residual(self.sol_opts["subtract-model"], slice(None))
                resid_vis = np.zeros_like(resid_vis1)
                self.gm.apply_inv_gains(resid_vis1, resid_vis, full2x2=True,
                                        direction=self.sol_opts["correct-dir"])
                del resid_vis1

            # clear the SKIPSOL flag to also flag data that's been omitted from the solutions
            self.vdm.flags_arr &= ~FL.SKIPSOL

            # run madmax on them
            self.madmax.set_mode(GD['madmax']['residuals'])
            thr1, thr2 = self.madmax.get_mad_thresholds()
            if thr1 or thr2:
                if self.madmax.beyond_thunderdome(resid_vis, None, None, self.vdm.flags_arr, thr1, thr2,
                                             "{} residual".format(self.label)):
                    self.stats.chunk.num_mad_flagged = ((self.vdm.flags_arr & FL.MAD) != 0).sum()
            resid_vis = None  # release memory if new object was created

        # collect messages from various flagging sources, and print to log if any
        flagstatus = []

        if self.stats.chunk.num_sol_flagged:
            # also for up message with flagging stats
            fstats = []
            for flagname, mask in FL.categories().items():
                if mask != FL.MISSING:
                    n_flag, n_tot = self.gm.num_gain_flags(mask, final=True)
                    if n_flag:
                        fstats.append("{}:{}({:.2%})".format(flagname, n_flag, n_flag/float(n_tot)))

            nfl, nsol = self.gm.num_gain_flags(final=True)
            flagstatus.append("gain flags {} ({:.2%} total)".format(" ".join(fstats), nfl/float(nsol)))

        if self.stats.chunk.num_mad_flagged:
            flagstatus.append("MadMax took out {} visibilities ({} in final round)".format(
                        self.stats.chunk.num_mad_flagged, self.stats.chunk.num_mad_flagged - num_mad_flagged_prior))

        if flagstatus:
            n_new_flags = (self.vdm.flags_arr&~(FL.MISSING|FL.SKIPSOL) != 0).sum() - self.stats.chunk.num_prior_flagged
            if n_new_flags < self.vdm.flags_arr.size*GD['flags']['warn-thr']:
                warning, color = "", "blue"
            else:
                warning, color = "", "red"
            log(0, color).print("{}{} has {} ({:.2%}) new data flags: {}".format(
                warning, self.label,
                n_new_flags, n_new_flags / float(self.vdm.flags_arr.size),
                ", ".join(flagstatus)))


class SolveOnly(SolverMachine):
    """Runs the solver, but does not apply solutions"""

    def run(self):
#        import ipdb; ipdb.set_trace()

        _solve_gains(self.gm, self.stats, self.madmax,
                     self.vdm.weighted_obser, self.vdm.weighted_model, self.vdm.flags_arr,
                     self.sol_opts, label=self.label)

        self.output_weights = self.gm.output_weights

        if ifrgain_machine.is_computing():
            ifrgain_machine.update(self.vdm.weighted_obser, self.vdm.corrupt_weighted_model, self.vdm.flags_arr,
                                   self.vdm.freq_slice, self.soldict)

        return None



class SolveAndCorrect(SolveOnly):
    """
    Run the solver and save and apply the resulting gain solutions to the observed data. Produces
    corrected data.
    """
    def run(self):
        SolveOnly.run(self)
        # for corrected visibilities, take the first data/model pair only
        corr_vis = np.zeros_like(self.vdm.obser_arr)
        self.gm.apply_inv_gains(self.vdm.obser_arr, corr_vis, full2x2=True, direction=self.sol_opts["correct-dir"])
        return corr_vis

class SolveAndSubtract(SolveOnly):
    """
    Run the solver, generate residuals, and (optionally) apply the resulting gain solutions to the residuals.
    Produces (un)corrected residuals.
    """
    # The SolveAndCorrectResiduals subclass redefines this to True
    output_corrected_residuals = False

    def run(self, correct=False):
        SolveOnly.run(self)
        # compute residuals
        resid_vis = self.vdm.corrupt_residual(self.sol_opts["subtract-model"],  self.sol_opts["subtract-dirs"])

        # correct residual if required
        if self.output_corrected_residuals:
            corr_vis = np.zeros_like(resid_vis)
            self.gm.apply_inv_gains(resid_vis, corr_vis, full2x2=True, direction=self.sol_opts["correct-dir"])
            return corr_vis
        else:
            return resid_vis

class SolveAndCorrectResiduals(SolveAndSubtract):
    """
    Run the solver, generate corrected residuals.
    """
    # mark this machine as generating corrected residuals
    output_corrected_residuals = True
    # mark this machine as generating full corrected residuals, if all directions are subtracted
    @property
    def outputs_full_corrected_residuals(self):
        return self.sol_opts['subtract-dirs'] == slice(None)

class CorrectOnly(SolverMachine):
    """
    Do not solve. Apply prior gain solutions to the observed data, generating corrected data.
    """
    # mark machine as an apply-only type
    is_apply_only = True
    # model not required, unless we're flagging on residuals
    @classproperty
    def is_model_required(cls):
        return bool(GD['madmax']['residuals'])

    def run(self):
        if self.vdm.model_arr is not None and ifrgain_machine.is_computing():
            ifrgain_machine.update(self.vdm.weighted_obser, self.vdm.corrupt_weighted_model, self.vdm.flags_arr,
                                   self.vdm.freq_slice, self.soldict)

        corr_vis = np.zeros_like(self.vdm.obser_arr)
        self.gm.apply_inv_gains(self.vdm.obser_arr, corr_vis, full2x2=True, direction=self.sol_opts["correct-dir"])

        return corr_vis

class SubtractOnly(SolverMachine):
    """
    Do not solve. Apply prior gain solutions, generate (un)corrected residuals.
    The optional flag to run is invoked by the subclass
    """
    # mark machine as an apply-only type
    is_apply_only = True
    # The SubtractAndCorrect subclass redefines this to True
    output_corrected_residuals = False

    def run(self):
        # doing it before recomputing the residuals: saves time
        if ifrgain_machine.is_computing():
            ifrgain_machine.update(self.vdm.weighted_obser, self.vdm.corrupt_weighted_model, self.vdm.flags_arr,
                                   self.vdm.freq_slice, self.soldict)

        resid_vis = self.vdm.corrupt_residual(self.sol_opts["subtract-model"],  self.sol_opts["subtract-dirs"])

        # correct residual if required
        if self.output_corrected_residuals:
            corr_vis = np.zeros_like(resid_vis)
            self.gm.apply_inv_gains(resid_vis, corr_vis, full2x2=True, direction=self.sol_opts["correct-dir"])
            return corr_vis
        else:
            return resid_vis

class CorrectResiduals(SubtractOnly):
    """
    Do not solve. Apply prior gain solutions, generate corrected residuals.
    """
    # mark this machine as generating corrected residuals
    output_corrected_residuals = True
    # mark this machine as generating full corrected residuals, if all directions are subtracted
    @property
    def outputs_full_corrected_residuals(self):
        return self.sol_opts['subtract-dirs'] == slice(None)


SOLVERS = { 'so': SolveOnly,
            'sc': SolveAndCorrect,
            'sr': SolveAndCorrectResiduals,
            'ss': SolveAndSubtract,
            'ac': CorrectOnly,
            'ar': CorrectResiduals,
            'as': SubtractOnly
            }

#@profile
def run_solver(solver_type, itile, chunk_key, sol_opts, debug_opts):
    """
    Initialises a gain machine and invokes the solver for the current chunk.

    Args:
        solver_type (str):
            Specifies type of solver to use.
        itile (int):
            Index of current Tile object.
        chunk_key (str):
            Label identifying the current chunk (e.g. "D0T1F2").
        sol_opts (dict):
            Solver options (see [sol] section in DefaultParset.cfg).

    Returns:
        :obj:`~cubical.statistics.SolverStats`:
            An object containing solver statistics.

    Raises:
        RuntimeError:
            If gain factory has not been initialised.
    """
    import cubical.workers
    cubical.workers._init_worker()

    label = None
    try:
        tile = cubical.workers.tile_list[itile]
        label = chunk_key
        solver = SOLVERS[solver_type]
        # initialize the gain machine for this chunk

        if gm_factory is None:
            raise RuntimeError("Gain machine factory has not been initialized")

        # Get chunk data from tile.

        # need to know which kernel to use to allocate visibility and flag arrays
        allocate_vis_array, allocate_flag_array, _ = gm_factory.determine_allocators()

        obser_arr, model_arr, flags_arr, weight_arr = tile.get_chunk_cubes(chunk_key,
                                 gm_factory.ctype,
                                 allocator=allocate_vis_array,
                                 flag_allocator=allocate_flag_array)

        chunk_ts, chunk_fs, _, freq_slice = tile.get_chunk_tfs(chunk_key)

        # apply IFR-based gains, if any
        ifrgain_machine.apply(obser_arr, freq_slice)

        # create subdict in shared dict for solutions etc.
        soldict = tile.create_solutions_chunk_dict(chunk_key)

        # create VisDataManager for this chunk

        vdm = _VisDataManager(obser_arr, model_arr, flags_arr, weight_arr, freq_slice)

        n_dir, n_mod = model_arr.shape[0:2] if model_arr is not None else (1,1)

        solver_machine_class = SOLVERS[solver_type]
        # create GainMachine
        # import pudb; pu.db
        gm = vdm.gm = gm_factory.create_machine(vdm.weighted_obser, n_dir, n_mod, chunk_ts, chunk_fs, label)

        # create solver machine
        solver_machine = solver_machine_class(vdm, gm, soldict, sol_opts, label, metadata)

        # Invoke solver method
        if debug_opts['stop-before-solver']:
            import pdb
            pdb.set_trace()

        corr_vis = solver_machine.run()
        
        # Panic if amplitude has gone crazy

        if debug_opts['panic-amplitude']:
            if corr_vis is not None:
                unflagged = flags_arr==0
                if unflagged.any() and abs(corr_vis[unflagged,:,:]).max() > debug_opts['panic-amplitude']:
                    raise RuntimeError("excessive amplitude in chunk {}".format(label))

        # finalize residuals
        solver_machine.finalize(corr_vis)

        # Copy results back into tile.
        have_new_flags = (solver_machine.stats.chunk.num_sol_flagged > 0 or
                          solver_machine.stats.chunk.num_mad_flagged > 0)

        tile.set_chunk_cubes(corr_vis, flags_arr if have_new_flags else None,
                             solver_machine.output_weights, chunk_key)

        # Ask the gain machine to store its solutions in the shared dict.
        gm_factory.export_solutions(gm, soldict)

        # Trigger garbage collection because it seems very unreliable. This 
        # flattens the memory profile substantially. 
        gc.collect()

        return solver_machine.stats

    except Exception as exc:
        log.error("Solver for tile {} chunk {} failed with exception: {}".format(itile, label, exc))
        log.print(traceback.format_exc())
        raise




# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details


# This module has been adapted from the DDFacet package,
# (c) Cyril Tasse et al., see http://github.com/saopicc/DDFacet

#import sharedarray.SharedArray as SharedArray
from __future__ import print_function
from builtins import range
import SharedArray
from . import ModColor
from . import logger
import traceback
log = logger.getLogger("NpShared")
import os.path


def zeros(Name, *args, **kwargs):
    try:
        return SharedArray.create(Name, *args, **kwargs)
    except:
        DelArray(Name)
        return SharedArray.create(Name, *args, **kwargs)


def SizeShm():
    L = ListNames()
    S = 0
    for l in L:
        A = GiveArray(l)
        if A is not None:
            S += A.nbytes
    return float(S)/(1024**2)


def CreateShared(Name, shape, dtype):
    try:
        a = SharedArray.create(Name, shape, dtype=dtype)
    except OSError:
        print(ModColor.Str("File %s exists, deleting" % Name), file=log)
        DelArray(Name)
        a = SharedArray.create(Name, shape, dtype=dtype)
    return a

def ToShared(Name, A):

    a = CreateShared(Name, A.shape, A.dtype)
    a[:] = A[:]
    return a

def DelArray(Name):
    try:
        SharedArray.delete(Name)
    except:
        pass

_locking = True

def Lock (array):
    global _locking
    if _locking:
        try:
                SharedArray.mlock(array)
        except:
            print("Warning: Cannot lock memory. Try updating your kernel security settings.", file=log)
            _locking = False

def Unlock (array):
    global _locking
    if _locking:
        try:
                SharedArray.munlock(array)
        except:
            print("Warning Cannot unlock memory. Try updating your kernel security settings.", file=log)
            _locking = False


def ListNames():
    ll = list(SharedArray.list())
    return [AR.name for AR in ll]


def DelAll(key=None):
    ll = ListNames()
    for name in ll:
        if key is not None:
            if key in name:
                DelArray(name)
        else:
            DelArray(name)


def GiveArray(Name):
    # return SharedArray.attach(Name)
    try:
        return SharedArray.attach(Name)
    except Exception as e:  # as exception:
        # #print str(e)
        # print
        # print "Exception for key [%s]:"%Name
        # print "   %s"%(str(e))
        # print
        print("Error loading",Name)
        traceback.print_exc()
        return None


def Exists(Name):
    if Name.startswith("file://"):
        return os.path.exists(Name[7:])
    if Name.startswith("shm://"):
        Name = Name[6:]
    return Name in ListNames()


def DicoToShared(Prefix, Dico, DelInput=False):
    DicoOut = {}
    print(ModColor.Str("DicoToShared: start [prefix = %s]" % Prefix), file=log)
    for key in list(Dico.keys()):
        if not isinstance(Dico[key], np.ndarray):
            continue
        # print "%s.%s"%(Prefix,key)
        ThisKeyPrefix = "%s.%s" % (Prefix, key)
        print(ModColor.Str("  %s -> %s" % (key, ThisKeyPrefix)), file=log)
        ar = Dico[key]
        Shared = ToShared(ThisKeyPrefix, ar)
        DicoOut[key] = Shared
        if DelInput:
            del(Dico[key], ar)

    if DelInput:
        del(Dico)

    print(ModColor.Str("DicoToShared: done"), file=log)
    #print ModColor.Str("DicoToShared: done")

    return DicoOut


def SharedToDico(Prefix):

    print(ModColor.Str("SharedToDico: start [prefix = %s]" % Prefix), file=log)
    Lnames = ListNames()
    keys = [Name for Name in Lnames if Prefix in Name]
    if len(keys) == 0:
        return None
    DicoOut = {}
    for Sharedkey in keys:
        key = Sharedkey.split(".")[-1]
        print(ModColor.Str("  %s -> %s" % (Sharedkey, key)), file=log)
        Shared = GiveArray(Sharedkey)
        if isinstance(Shared, type(None)):
            print(ModColor.Str("      None existing key %s" % (key)), file=log)
            return None
        DicoOut[key] = Shared
    print(ModColor.Str("SharedToDico: done"), file=log)

    return DicoOut

####################################################
####################################################


def PackListArray(Name, LArray):
    DelArray(Name)

    NArray = len(LArray)
    ListNDim = [len(LArray[i].shape) for i in range(len(LArray))]
    NDimTot = np.sum(ListNDim)
    # [NArray,NDim0...NDimN,shape0...shapeN,Arr0...ArrN]

    dS = LArray[0].dtype
    TotSize = 0
    for i in range(NArray):
        TotSize += LArray[i].size

    S = SharedArray.create(Name, (1+NArray+NDimTot+TotSize,), dtype=dS)
    S[0] = NArray
    idx = 1
    # write ndims
    for i in range(NArray):
        S[idx] = ListNDim[i]
        idx += 1

    # write shapes
    for i in range(NArray):
        ndim = ListNDim[i]
        A = LArray[i]
        S[idx:idx+ndim] = A.shape
        idx += ndim

    # write arrays
    for i in range(NArray):
        A = LArray[i]
        S[idx:idx+A.size] = A.ravel()
        idx += A.size


def UnPackListArray(Name):
    S = GiveArray(Name)

    NArray = np.int32(S[0].real)
    idx = 1

    # read ndims
    ListNDim = []
    for i in range(NArray):
        ListNDim.append(np.int32(S[idx].real))
        idx += 1

    # read shapes
    ListShapes = []
    for i in range(NArray):
        ndim = ListNDim[i]
        shape = np.int32(S[idx:idx+ndim].real)
        ListShapes.append(shape)
        idx += ndim

    # read values
    ListArray = []
    for i in range(NArray):
        shape = ListShapes[i]
        size = np.prod(shape)
        A = S[idx:idx+size].reshape(shape)
        ListArray.append(A)
        idx += size
    return ListArray

####################################################
####################################################


def PackListSquareMatrix(shared_dict, Name, LArray):

    NArray = len(LArray)
    dtype = LArray[0].dtype
    TotSize = 0
    for i in range(NArray):
        TotSize += LArray[i].size

    # [N,shape0...shapeN,Arr0...ArrN]
    S = shared_dict.addSharedArray(Name, (TotSize+NArray+1,), dtype=dtype)
    S[0] = NArray
    idx = 1
    for i in range(NArray):
        A = LArray[i]
        S[idx] = A.shape[0]
        idx += 1

    for i in range(NArray):
        A = LArray[i]
        S[idx:idx+A.size] = A.ravel()
        idx += A.size


def UnPackListSquareMatrix(Array):
    LArray = []
    S = GiveArray(Array) if type(Array) is str else Array

    NArray = np.int32(S[0].real)
    idx = 1

    ShapeArray = []
    for i in range(NArray):
        ShapeArray.append(np.int32(S[idx].real))
        idx += 1

    print(ShapeArray, file=log)

    for i in range(NArray):
        shape = np.int32(ShapeArray[i].real)
        size = shape**2
        A = S[idx:idx+size].reshape((shape, shape))
        LArray.append(A)
        idx += A.size
    return LArray


# import SharedArray
# import ModColor

# def ToShared(Name,A):

#     try:
#         a=SharedArray.create(Name,A.shape,dtype=A.dtype)
#     except:
#         print ModColor.Str("File %s exists, delete it..."%Name)
#         DelArray(Name)
#         a=SharedArray.create(Name,A.shape,dtype=A.dtype)


#     a[:]=A[:]
#     return a

# def DelArray(Name):
#     SharedArray.delete(Name)

# def GiveArray(Name):
#     return SharedArray.attach(Name)
